from mhcgnomes import Allele, Species, parse

from .common import eq_


def test_parse_RT1():
    species = parse("RT1")
    eq_(type(species), Species)
    eq_(species.prefix, "RT1")


def test_Species_get_RT1():
    species = Species.get("RT1")
    eq_(type(species), Species)
    eq_(species.prefix, "RT1")


def test_parse_Rano():
    species = parse("Rano")
    eq_(type(species), Species)
    eq_(species.prefix, "Rano")


def test_Species_get_Rano():
    species = Species.get("Rano")
    eq_(type(species), Species)
    eq_(species.prefix, "Rano")


def test_rat_class1_allele_Bbu():
    eq_(parse("RT1-Bb*u"), Allele.get("RT1", "Bb", "u"))


def test_rat_class1_allele_Bbu_no_star():
    eq_(parse("RT1-Bbu"), Allele.get("RT1", "Bb", "u"))


def test_rat_class1_allele_Db1a():
    eq_(parse("RT1-Db1*a"), Allele.get("RT1", "Db1", "a"))


def test_rat_class1_allele_Db1a_no_star():
    eq_(parse("RT1-Db1a"), Allele.get("RT1", "Db1", "a"))


def test_rat_class1_allele_DMaa():
    eq_(parse("RT1-DMa*a"), Allele.get("RT1", "DMA", "a"))


def test_rat_class1_allele_DMaa_no_star():
    eq_(parse("RT1-DMAa"), Allele.get("RT1", "DMA", "a"))


def test_rat_class1_allele_95f():
    eq_(parse("RT1-9.5*f"), Allele.get("RT1", "9.5", "f"))


def test_rat_class1_allele_95f_no_star():
    eq_(parse("RT1-9.5f"), Allele.get("RT1", "9.5", "f"))


def test_rat_class1_allele_M3_1_av1():
    expected = Allele.get("RT1", "M3-1", "av1")
    assert expected is not None
    eq_(parse("RT1-M3-1*av1"), expected)


def test_rat_class1_allele_M3_1_av1_no_star():
    expected = Allele.get("RT1", "M3-1", "av1")
    assert expected is not None
    eq_(parse("RT1-M3-1av1"), expected)


"""
def test_RT1_Db_A_weird_uniprot_name():
    seq = "Class II RT1.D(A) beta chain"
    expected = Allele.get("RT1", "Db", "a")
    assert expected is not None
    eq_(parse(seq), expected)
"""



from mhcgnomes import Allele, Mutation

from .common import eq_


def test_allele_get_parses_mutation_strings():
    allele = Allele.get("HLA", "A", "02", "01", mutations="N80I")
    expected = Mutation.get(pos=80, aa_original="N", aa_mutant="I")
    eq_(allele.mutations, (expected,))
    assert allele.is_mutant
    assert allele.to_string().endswith("N80I mutant")


def test_mutation_get_preserves_raw_string():
    mutation = Mutation.get(pos=86, aa_original="G", aa_mutant="Y", raw_string="G86Y")
    eq_(mutation.raw_string, "G86Y")



from dataclasses import FrozenInstanceError

import pytest

from mhcgnomes import Species, parse
from mhcgnomes.tokenize import tokenize


def test_parse_returns_immutable_cached_result():
    result1 = parse("HLA-A*02:01")
    result2 = parse("HLA-A*02:01")

    assert result1 is result2

    with pytest.raises(FrozenInstanceError):
        result1.raw_string = "MUTATED"
    with pytest.raises(FrozenInstanceError):
        result1.gene.name = "Z"

    result3 = parse("HLA-A*02:01")
    assert result3.raw_string == "HLA-A*02:01"
    assert result3.gene.name == "A"
    assert result3.to_string() == "HLA-A*02:01"


def test_tokenize_returns_immutable_cached_result():
    tokenization1 = tokenize("HLA-A*02:01")
    tokenization2 = tokenize("HLA-A*02:01")

    assert tokenization1 is tokenization2

    with pytest.raises(TypeError):
        tokenization1.attributes["BAD"] = "1"
    with pytest.raises(FrozenInstanceError):
        tokenization1.tokens[0].raw_string = "BAD"

    tokenization3 = tokenize("HLA-A*02:01")
    assert "BAD" not in tokenization3.attributes
    assert tokenization3.tokens[0].raw_string != "BAD"
    assert tokenization3.tokens == tokenize("HLA-A*02:01").tokens


def test_species_get_multiple_returns_immutable_cached_collection():
    matches = Species.get_multiple("HLA")

    assert isinstance(matches, tuple)
    assert all(match.prefix == "HLA" for match in matches)
    with pytest.raises(TypeError):
        matches[0].gene_names.add("BAD")

    modified = (*matches, "BAD")
    assert modified[-1] == "BAD"
    assert all(match.prefix == "HLA" for match in Species.get_multiple("HLA"))



from mhcgnomes import dataframe_from_parsed_objects, dataframe_from_string_list, parse


def test_dataframe_from_parsed_objects_uses_to_record_output():
    df = dataframe_from_parsed_objects([parse("HLA-A*02:01"), parse("HLA-B*07:02")])

    assert df["gene"].tolist() == ["HLA-A", "HLA-B"]
    assert df["allele"].tolist() == ["HLA-A*02:01", "HLA-B*07:02"]
    assert df["mhc_class"].tolist() == ["Ia", "Ia"]


def test_dataframe_from_string_list_parses_names_before_dataframe_conversion():
    df = dataframe_from_string_list(["HLA-A*02:01", "HLA-B*07:02"])

    assert df["species_prefix"].tolist() == ["HLA", "HLA"]
    assert df["gene"].tolist() == ["HLA-A", "HLA-B"]
    assert df["allele"].tolist() == ["HLA-A*02:01", "HLA-B*07:02"]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ParseError(Exception):
    """Raised when an MHC string cannot be parsed."""

    pass


class UnknownSpecies(ParseError):
    pass


class UnknownGene(ParseError):
    pass


class UnknownHaplotype(ParseError):
    pass


class UnknownSerotype(ParseError):
    pass



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import OrderedDict, defaultdict
from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from types import MappingProxyType
from typing import Any, Union

from .common import cache
from .data import allele_aliases as raw_allele_aliases_dict
from .data import gene_aliases as raw_gene_aliases_dict
from .data import haplotypes as raw_haplotypes_dict
from .data import heterodimers as raw_heterodimers_dict
from .data import known_alleles as raw_known_alleles_dict
from .data import serotypes as raw_serotypes_dict
from .data import species as raw_species_dict
from .data import supertypes as raw_supertypes_dict
from .mhc_class_helpers import class1_restrictions, class2_restrictions
from .normalizing_dictionary import NormalizingDictionary
from .normalizing_set import NormalizingSet
from .result import Result


class FrozenList(list):
    def _immutable(self, *args, **kwargs):
        raise TypeError("FrozenList is immutable")

    append = _immutable
    clear = _immutable
    extend = _immutable
    insert = _immutable
    pop = _immutable
    remove = _immutable
    reverse = _immutable
    sort = _immutable
    __delitem__ = _immutable
    __iadd__ = _immutable
    __imul__ = _immutable
    __setitem__ = _immutable

    def copy(self):
        return list(self)


class FrozenSet(set):
    def _immutable(self, *args, **kwargs):
        raise TypeError("FrozenSet is immutable")

    add = _immutable
    clear = _immutable
    difference_update = _immutable
    discard = _immutable
    intersection_update = _immutable
    pop = _immutable
    remove = _immutable
    symmetric_difference_update = _immutable
    update = _immutable
    __ior__ = _immutable
    __iand__ = _immutable
    __isub__ = _immutable
    __ixor__ = _immutable

    def copy(self):
        return set(self)

    def __repr__(self):
        return repr(set(self))

    def __str__(self):
        return str(set(self))


def _ensure_normalizing_set(values):
    if values is None:
        return NormalizingSet()
    if isinstance(values, NormalizingSet):
        return values.copy()
    return NormalizingSet(*values)


def _ensure_normalizing_dictionary(values, default_value_fn=None):
    if values is None:
        return NormalizingDictionary(default_value_fn=default_value_fn)
    if isinstance(values, NormalizingDictionary):
        result = values.copy()
        if default_value_fn is not None and result.default_value_fn is None:
            result.default_value_fn = default_value_fn
        return result
    return NormalizingDictionary.from_dict(values, default_value_fn=default_value_fn)


def _freeze_nested_value(value):
    if isinstance(value, NormalizingDictionary):
        frozen_dict = value.copy()
        for key, subvalue in list(frozen_dict.items()):
            frozen_dict[key] = _freeze_nested_value(subvalue)
        return frozen_dict.freeze()
    if isinstance(value, NormalizingSet):
        return value.copy().freeze()
    if isinstance(value, dict):
        return MappingProxyType({k: _freeze_nested_value(v) for k, v in value.items()})
    if isinstance(value, (list, tuple)):
        return FrozenList(_freeze_nested_value(v) for v in value)
    if isinstance(value, (set, frozenset)):
        return FrozenSet(_freeze_nested_value(v) for v in value)
    return value


@dataclass(eq=False, repr=False, frozen=True, init=False)
class Species(Result):
    """
    Representation of a parsed species prefix such as "HLA", "ELA"
    """

    name: str = ""
    common_name: str = ""
    mhc_prefix: str = ""
    gene_names: Any = None
    gene_name_to_mhc_class: Any = None
    class2_loci: Any = None
    class2_locus_to_gene_names: Any = None
    class2_gene_name_to_chain_type: Any = None
    gene_aliases: Any = None
    reverse_gene_aliases: Any = None
    allele_aliases: Any = None
    known_alleles: Any = None
    haplotypes: Any = None
    serotypes: Any = None
    heterodimers: Any = None
    supertypes: Any = None
    parent_species: Union["Species", None] = None
    old_mhc_prefix: str = ""
    other_mhc_prefixes: Any = None
    other_common_names: Any = None

    def __init__(
        self,
        name: str,
        common_name: str,
        mhc_prefix: str,
        gene_names: Iterable[str],
        gene_name_to_mhc_class: Mapping[str, str],
        class2_loci: Iterable[str],
        class2_locus_to_gene_names: Mapping[str, Iterable[str]],
        class2_gene_name_to_chain_type: Mapping[str, str],
        gene_aliases: Mapping[str, str],
        allele_aliases: Mapping[str, str],
        known_alleles: Mapping[str, Iterable[str]],
        haplotypes: Mapping[str, Iterable[str]],
        serotypes: Mapping[str, Iterable[str]],
        heterodimers: Mapping[str, Mapping[str, str]],
        supertypes: Mapping[str, Mapping[str, Any]],
        parent_species: Union["Species", None] = None,
        old_mhc_prefix: Union[str, None] = None,
        other_mhc_prefixes: Iterable[str] = [],
        other_common_names: Iterable[str] = [],
        raw_string: Union[str, None] = None,
    ):
        Result.__init__(self, raw_string=raw_string)

        gene_names = _ensure_normalizing_set(gene_names)
        gene_name_to_mhc_class = _ensure_normalizing_dictionary(gene_name_to_mhc_class)
        class2_loci = _ensure_normalizing_set(class2_loci)
        class2_locus_to_gene_names = _ensure_normalizing_dictionary(
            class2_locus_to_gene_names, default_value_fn=set
        )
        class2_gene_name_to_chain_type = _ensure_normalizing_dictionary(
            class2_gene_name_to_chain_type
        )
        gene_aliases = _ensure_normalizing_dictionary(gene_aliases)
        allele_aliases = _ensure_normalizing_dictionary(allele_aliases)
        known_alleles = _ensure_normalizing_dictionary(
            known_alleles, default_value_fn=NormalizingSet
        )
        known_alleles = known_alleles.map_values(_ensure_normalizing_set)
        haplotypes = _ensure_normalizing_dictionary(haplotypes)
        serotypes = _ensure_normalizing_dictionary(serotypes)
        heterodimers = _ensure_normalizing_dictionary(heterodimers)
        supertypes = _ensure_normalizing_dictionary(supertypes)

        self._set_field(self, "name", name)
        self._set_field(self, "common_name", common_name)
        self._set_field(self, "other_common_names", FrozenList(other_common_names))
        self._set_field(self, "mhc_prefix", mhc_prefix)
        self._set_field(self, "other_mhc_prefixes", FrozenSet(other_mhc_prefixes))
        if old_mhc_prefix:
            normalized_old_mhc_prefix = old_mhc_prefix
        else:
            normalized_old_mhc_prefix = mhc_prefix
        self._set_field(self, "old_mhc_prefix", normalized_old_mhc_prefix)

        self._set_field(self, "gene_names", _freeze_nested_value(gene_names))
        self._set_field(
            self, "gene_name_to_mhc_class", _freeze_nested_value(gene_name_to_mhc_class)
        )
        self._set_field(self, "class2_loci", _freeze_nested_value(class2_loci))
        self._set_field(
            self,
            "class2_locus_to_gene_names",
            _freeze_nested_value(class2_locus_to_gene_names),
        )
        self._set_field(
            self,
            "class2_gene_name_to_chain_type",
            _freeze_nested_value(class2_gene_name_to_chain_type),
        )
        self._set_field(self, "gene_aliases", _freeze_nested_value(gene_aliases))
        # create a reverse lookup from proper names to their list of aliases
        reverse_gene_aliases = self._create_reverse_gene_aliases(self.gene_names, self.gene_aliases)
        self._set_field(self, "reverse_gene_aliases", _freeze_nested_value(reverse_gene_aliases))

        self._set_field(self, "allele_aliases", _freeze_nested_value(allele_aliases))
        self._set_field(self, "known_alleles", _freeze_nested_value(known_alleles))
        self._set_field(self, "haplotypes", _freeze_nested_value(haplotypes))
        self._set_field(self, "serotypes", _freeze_nested_value(serotypes))
        self._set_field(self, "heterodimers", _freeze_nested_value(heterodimers))
        self._set_field(self, "supertypes", _freeze_nested_value(supertypes))
        self._set_field(self, "parent_species", parent_species)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if type(other) is not Species:
            return False
        return self.name == other.name

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    @classmethod
    def _create_reverse_gene_aliases(cls, gene_names, gene_aliases):
        """
        Creates dictionary mapping canonical gene name to its set of aliases.
        """
        d = defaultdict(NormalizingSet)

        for k, v in gene_aliases.items():
            d[v].add(k)
        for v in gene_names:
            d[v].add(v)
        return d

    @classmethod
    def str_field_names(cls):
        return ("name", "mhc_prefix")

    @classmethod
    def tuple_field_names(cls):
        return ("name", "mhc_prefix")

    @classmethod
    def eq_field_names(cls):
        return ("name",)

    @classmethod
    def from_dict(cls, d):
        if "name" in d:
            species_name = d["name"]
            return species_name_to_species_object.get(species_name)
        return None

    @property
    def species_name(self):
        return self.name

    @property
    def canonical_mhc_prefix(self):
        return self.mhc_prefix

    @property
    def parent(self):
        return self.parent_species

    # TODO: If taxonomy-aware operations keep expanding, promote these lineage
    # helpers into a fuller Taxon API instead of growing Species ad hoc.
    def is_parent_of(self, other):
        other = Species.get(other)
        return other is not None and other.parent_species == self

    def is_child_of(self, other):
        other = Species.get(other)
        return other is not None and self.parent_species == other

    def is_ancestor_of(self, other):
        other = Species.get(other)
        if other is None:
            return False
        current = other.parent_species
        while current is not None:
            if current == self:
                return True
            current = current.parent_species
        return False

    def is_descendant_of(self, other):
        other = Species.get(other)
        if other is None:
            return False
        current = self.parent_species
        while current is not None:
            if current == other:
                return True
            current = current.parent_species
        return False

    @property
    def historic_mhc_prefix(self):
        """
        Return older species name which is now used to group multiple
        related species (e.g. "DLA" for "Calu").
        """
        return self.old_mhc_prefix

    @property
    def historic_alias(self):
        return self.historic_mhc_prefix

    @property
    def gene_names_and_aliases(self):
        name_set = set(self.gene_aliases.keys()).union(self.gene_names)
        return sorted(name_set)

    @property
    def num_genes(self):
        return len(self.gene_names)

    @property
    def scientific_species_name(self):
        return self.name

    @property
    def latin_name(self):
        return self.name

    @property
    def species_prefix(self):
        return self.canonical_mhc_prefix

    @property
    def prefix(self):
        return self.species_prefix

    def to_string(self, include_species=True, use_old_species_prefix=False):
        if not include_species:
            return ""
        elif use_old_species_prefix:
            return self.historic_alias
        else:
            return self.prefix

    def compact_string(self, include_species=False, use_old_species_prefix=False):
        return self.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )

    @classmethod
    @cache
    def get_multiple(cls, species_name):
        """
        Cached wrapper around find_matching_species which tries to
        identify one or more species from any of its scientific name,
        common names, MHC prefixes.
        """
        return tuple(find_matching_species_objects(species_name))

    @classmethod
    @cache
    def get(cls, species_name):
        """
        Look up a species by any identifier (latin name, prefix, common name).
        Returns None if no match or if the alias is ambiguous (maps to
        multiple species). Use get_multiple() to retrieve all candidates
        for an ambiguous alias.
        """
        if type(species_name) is Species:
            return species_name
        elif species_name is None or type(species_name) is not str:
            return None

        species_objects = cls.get_multiple(species_name)

        # Try underscore-to-space normalization if no results
        # (common in bioinformatics: "Canis_lupus" for "Canis lupus")
        if len(species_objects) == 0 and "_" in species_name:
            species_objects = cls.get_multiple(species_name.replace("_", " "))

        if len(species_objects) == 0:
            return None
        if len(species_objects) == 1:
            return species_objects[0]
        # Multiple species match — try to find an unambiguous winner:
        # 1. Exact latin name match
        for sp in species_objects:
            if sp.name == species_name:
                return sp
        # 2. Exact primary prefix match (only one species should own a prefix)
        prefix_matches = [
            sp
            for sp in species_objects
            if sp.prefix.lower().strip() == species_name.lower().strip()
        ]
        if len(prefix_matches) == 1:
            return prefix_matches[0]
        # 3. For auto-generated long prefixes (e.g., CanisLupus), prefer
        # the species that isn't a subspecies (no parent with same identifier)
        non_child = [sp for sp in species_objects if sp.parent_species is None]
        if len(non_child) == 1:
            return non_child[0]
        # Ambiguous alias: return None instead of picking a heuristic winner
        return None

    @classmethod
    @cache
    def get_by_latin_name(cls, latin_name):
        """
        Look up a species by its scientific (latin) name.
        This is the canonical identity lookup and is never ambiguous.
        """
        return latin_name_to_species_object.get(latin_name)

    def to_record(self):
        return OrderedDict(
            [
                ("species_prefix", self.prefix),
                ("species_name", self.name),
                ("species_latin_name", self.latin_name),
            ]
        )

    @property
    def common_species_name(self):
        """
        Returns common species name associated with MHC species
        prefix.
        """
        return self.common_name

    @property
    def all_mhc_prefixes(self):
        """
        Returns all MHC prefixes used for this species
        """
        prefixes = set(self.other_mhc_prefixes)
        if self.old_mhc_prefix and self.old_mhc_prefix != self.prefix:
            prefixes.add(self.old_mhc_prefix)
        return [self.prefix, *sorted(prefixes)]

    @property
    def all_common_names(self):
        """
        Returns all common names used for this species
        """
        return [self.common_name, *self.other_common_names]

    @property
    def all_identifiers(self):
        """
        Return all names and prefixes associated with this Species.

        Returns list of str
        """
        return [self.name, *self.all_mhc_prefixes, *self.all_common_names]

    def find_matching_gene_name(self, gene_name):
        """
        Use known aliases and normalized capitalization to infer
        the canonical gene name corresponding to the input.

        Returns str or None
        """
        if type(gene_name) in (int, float):
            gene_name = str(gene_name)
        if gene_name in self.gene_names:
            return self.gene_names.get_original(gene_name)
        elif gene_name in self.gene_aliases:
            alt_gene_name = self.gene_aliases[gene_name]
            if alt_gene_name in self.gene_names:
                # make sure the new gene name is a gene name
                # and not a locus
                # e.g. H2-IA -> H2-A is a mapping of loci
                return alt_gene_name
        # Strip ZFIN-style "mhc1"/"mhc2" prefix from gene tokens
        # (e.g. "mhc1uba" → "UBA", "mhc2daa" → "DAA"). This is a
        # convention used in zebrafish and potentially other fish species.
        lower = gene_name.lower()
        for prefix in ("mhc1", "mhc2"):
            if lower.startswith(prefix) and len(gene_name) > len(prefix):
                stripped = gene_name[len(prefix) :]
                if stripped in self.gene_names:
                    return self.gene_names.get_original(stripped)
        return None

    def find_matching_class2_locus_name(self, locus_name):
        """
        Use known aliases and normalized capitalization to infer
        the canonical gene name corresponding to the input.

        Returns str or None
        """
        if type(locus_name) in (int, float):
            locus_name = str(locus_name)
        if locus_name in self.class2_loci:
            return self.class2_loci.get_original(locus_name)
        elif locus_name in self.gene_aliases:
            alt_locus_name = self.gene_aliases[locus_name]
            if alt_locus_name in self.class2_loci:
                # make sure the new locus name is an actual locus
                # and not a gene
                # e.g.
                # H2-IA -> H2-A is a mapping of loci
                # but
                # H2-IAb -> H2-Ab is a mppaing of genes
                return alt_locus_name
        return None

    def normalize_gene_name_if_exists(self, gene_name):
        normalized_name = self.find_matching_gene_name(gene_name)

        if normalized_name:
            return normalized_name
        else:
            return gene_name

    def get_mhc_class_of_gene(self, gene_name):
        """
        Parameters
        ----------
        gene_name : str

        Returns either one of:
            {"I", "Ia", "Ib", "Ic", "Id", "II", IIa", "IIb", "other"}
        or None if gene can't be found
        """
        gene_name = self.normalize_gene_name_if_exists(gene_name)
        return self.gene_name_to_mhc_class.get(gene_name)

    def get_known_allele(self, gene_name, allele_name):
        gene_name_candidates = {gene_name}

        if gene_name is not None:
            gene_aliases = self.reverse_gene_aliases.get(gene_name, [])
            gene_name_candidates.update(gene_aliases)

        for gene_name in gene_name_candidates:
            known_alleles_for_gene_name = self.known_alleles.get(gene_name)
            if known_alleles_for_gene_name and allele_name in known_alleles_for_gene_name:
                known_allele_name = known_alleles_for_gene_name.get_original(allele_name)
                return (gene_name, known_allele_name)

        # if allele isn't in known_alleles but it's an alias for an allele which
        # is, then also return it
        for gene_name in gene_name_candidates:
            if gene_name:
                key = f"{gene_name}*{allele_name}"
            else:
                key = allele_name
            if key in self.allele_aliases:
                new_name = self.allele_aliases[key]
                for gene_name2 in gene_name_candidates:
                    if new_name in self.known_alleles.get(gene_name2, []):
                        known_allele_name = self.allele_aliases.original_key(key)
                        return (gene_name, known_allele_name)
        return None

    @cache
    def get_allele_alias(self, gene_name: str, allele_name: str):
        """
        Returns None if no alias found, otherwise pair of
        (gene name, allele name)
        """
        gene_name_candidates = {gene_name}

        if gene_name is not None:
            gene_aliases = self.reverse_gene_aliases.get(gene_name, [])
            gene_name_candidates.update(gene_aliases)

        # if allele isn't in known_alleles try looking at the keys of
        # allele_aliases
        for gene_name in gene_name_candidates:
            if gene_name:
                key = f"{gene_name}*{allele_name}"
            else:
                key = allele_name
            if key in self.allele_aliases:
                value = self.allele_aliases[key]
                if "*" in value:
                    parts = value.split("*")
                    if len(parts) != 2:
                        continue
                    return tuple(parts)
                else:
                    return (None, value)
        return None

    @classmethod
    def get_species_with_gene_name(self, gene_name):
        """
        Returns list of Species which have the given gene name
        """
        if gene_name in gene_name_to_species_objects:
            return list(gene_name_to_species_objects[gene_name])
        else:
            return []

    @property
    def is_mouse(self):
        return self.prefix in {"H-2", "H2"}

    @property
    def is_chicken(self):
        return self.prefix == "Gaga"

    @property
    def is_rat(self):
        return self.prefix in {"RT1", "Rano"}

    @property
    def is_human(self):
        return self.prefix == "HLA"

    @property
    def is_dog(self):
        return self.prefix in {"DLA", "Calu"}

    @property
    def is_cat(self):
        return self.prefix in {"FLA", "Feca"}

    @property
    def is_pig(self):
        return self.prefix in {"SLA", "Susc"}

    @property
    def is_cow(self):
        return self.prefix == "BoLA"

    @property
    def is_horse(self):
        return self.prefix in {"ELA", "Eqca"}


################################################################################
#
# Parse the species/gene ontology and create dictionaries mapping
# to Species objects
#
################################################################################


def guess_class2_chain_type(gene_name):
    # For now we're guessing based on the name, e.g.
    # DRB1 is a beta chain and DRA is an alpha chain
    trimmed_gene_name = gene_name
    if trimmed_gene_name.endswith("like"):
        trimmed_gene_name = trimmed_gene_name[:-4]
    while trimmed_gene_name and trimmed_gene_name[-1].isdigit():
        trimmed_gene_name = trimmed_gene_name[:-1]
    if not trimmed_gene_name:
        return "beta"
    last_letter = trimmed_gene_name[-1].upper()
    is_alpha = last_letter == "A"
    # assume that anything which can't be pinned down to be an alpha chain
    # is then a beta chain, since more of the variability/gene copying
    # seems to occur in beta chains
    return "alpha" if is_alpha else "beta"


def _make_long_prefix(latin_name):
    """
    Generate a 4+4 novel prefix from a latin binomial name by taking the first
    4 characters of each word, capitalized. E.g.:
        "Oryzias latipes" → "OryzLati"
        "Struthio camelus" → "StruCame"
        "Bubo bubo" → "BuboBubo"

    Returns None for names that don't have at least two words (e.g. "Bos sp.").
    """
    parts = latin_name.split()
    if len(parts) < 2 or parts[1].endswith("."):
        return None
    genus = parts[0][:4]
    species = parts[1][:4]
    return genus.capitalize() + species.capitalize()


def _make_5_5_prefix(latin_name):
    """
    Generate a 5+5 long prefix from a latin binomial name by taking the first
    5 characters of each word, capitalized. Kept for backward compatibility.
    """
    parts = latin_name.split()
    if len(parts) < 2 or parts[1].endswith("."):
        return None
    genus = parts[0][:5]
    species = parts[1][:5]
    return genus.capitalize() + species.capitalize()


def _is_taxonomic_prefix(prefix, latin_name):
    """
    Taxonomic helper prefixes like "Galliformes" or "Crocodylia" should not
    become inherited old MHC prefixes for child species.
    """
    normalized_prefix = re.sub(r"[^A-Za-z0-9]+", "", prefix).lower()
    latin_parts = [part for part in latin_name.split() if part.lower() != "sp."]
    if not latin_parts:
        return False
    normalized_first = re.sub(r"[^A-Za-z0-9]+", "", latin_parts[0]).lower()
    normalized_concat = "".join(re.sub(r"[^A-Za-z0-9]+", "", part).lower() for part in latin_parts)
    return normalized_prefix in {normalized_first, normalized_concat}


@cache
def create_species_for_latin_name(latin_name):
    if latin_name not in raw_species_dict:
        raise ValueError(f"Species not found: '{latin_name}'")
    species_info = raw_species_dict[latin_name]
    parent_species_latin_name = species_info.get("parent")
    if parent_species_latin_name:
        parent_species = create_species_for_latin_name(parent_species_latin_name)
    else:
        parent_species = None

    prefix = species_info.get("prefix")
    if not prefix:
        raise ValueError(f"Missing 'prefix' for '{latin_name}' in species ontology")

    old_mhc_prefix = species_info.get("old prefix")
    if not old_mhc_prefix and parent_species:
        parent_prefix = parent_species.prefix
        if not _is_taxonomic_prefix(parent_prefix, parent_species.name):
            old_mhc_prefix = parent_prefix

    other_mhc_prefixes = species_info.get("other prefixes")
    if type(other_mhc_prefixes) is str:
        other_mhc_prefixes = [other_mhc_prefixes]
    elif other_mhc_prefixes is None:
        other_mhc_prefixes = []

    # Auto-generate parseable aliases from the latin name:
    # 1. A 4+4 novel prefix (e.g. "OryzLati") — the standard for non-literature prefixes
    # 2. A 5+5 long prefix (e.g. "OryziLatip") for backward compatibility
    # 3. The full concatenated latin name (e.g. "OryziasLatipes") — always collision-free
    long_prefix = _make_long_prefix(latin_name)
    compat_prefix = _make_5_5_prefix(latin_name)
    parts = latin_name.split()
    full_concat = parts[0].capitalize() + parts[1].capitalize() if len(parts) >= 2 else None
    for alias in [long_prefix, compat_prefix, full_concat]:
        if alias and alias not in other_mhc_prefixes:
            other_mhc_prefixes = [*list(other_mhc_prefixes), alias]

    common_name = species_info.get("name")
    if not common_name:
        raise ValueError(f"Missing 'name' for '{latin_name}' in species ontology")

    if type(common_name) is str:
        common_names = [common_name]
    else:
        common_names = common_name
    # make all common names lowercase
    common_names = [s.lower() for s in common_names]
    shortest_common_name = min(common_names, key=len)

    if parent_species is None:
        gene_names = NormalizingSet()
        gene_name_to_mhc_class = NormalizingDictionary()
        class2_loci = NormalizingSet()
        class2_locus_to_gene_names = NormalizingDictionary(default_value_fn=set)
        class2_gene_name_to_chain_type = NormalizingDictionary()
    else:
        gene_names = parent_species.gene_names.copy()
        gene_name_to_mhc_class = parent_species.gene_name_to_mhc_class.copy()
        class2_loci = parent_species.class2_loci.copy()
        class2_locus_to_gene_names = parent_species.class2_locus_to_gene_names.map_values(set)
        class2_gene_name_to_chain_type = parent_species.class2_gene_name_to_chain_type.copy()

    for mhc_class, mhc_class_members in species_info.get("genes", {}).items():
        if mhc_class_members is None:
            raise ValueError(
                f"Unexpected None in gene ontology for class '{mhc_class}' of '{latin_name}'"
            )
        if mhc_class in class1_restrictions.union({"other"}):
            if type(mhc_class_members) is not list:
                raise ValueError(
                    f"Malformed gene ontology for '{latin_name}' MHC class '{mhc_class}'"
                )
            for gene_name in mhc_class_members:
                gene_names.add(str(gene_name))
                gene_name_to_mhc_class[gene_name] = mhc_class
        elif mhc_class in class2_restrictions:
            if type(mhc_class_members) is not dict:
                raise ValueError(
                    f"Malformed gene ontology for '{latin_name}' MHC class '{mhc_class}'"
                )
            for locus, locus_gene_names in mhc_class_members.items():
                class2_loci.add(locus)
                for gene_name in locus_gene_names:
                    gene_name = str(gene_name)
                    gene_names.add(gene_name)
                    gene_name_to_mhc_class[gene_name] = mhc_class
                    class2_locus_to_gene_names[locus].add(gene_name)
                    # TODO:
                    #  make alpha vs. beta chain genes explicit in
                    #  the YAML file ontology
                    chain_type = guess_class2_chain_type(gene_name)
                    class2_gene_name_to_chain_type[gene_name] = chain_type

    # Side tables inherit by lineage, but parent prefixes/common names must not
    # become implicit aliases for child species.
    ancestor_latin_names = []
    current_parent = parent_species
    while current_parent is not None:
        ancestor_latin_names.append(current_parent.name)
        current_parent = current_parent.parent_species

    all_identifiers = list(reversed(ancestor_latin_names))
    all_identifiers.append(latin_name)
    all_identifiers.append(prefix)
    if old_mhc_prefix and old_mhc_prefix != prefix:
        all_identifiers.append(old_mhc_prefix)
    all_identifiers.extend(other_mhc_prefixes)
    all_identifiers.extend(common_names)

    # Don't rely on other YAML files to use any particular prefix or
    # species identifier, just try them all
    # One reason for this is that canonical species prefixes might change
    # over time and don't want one update to break code mysteriously
    # elsewhere. Another reason is that a few species share the same
    # prefix ('Bubu' belongs to both an owl species and water buffalo).
    gene_aliases = combine_species_aliases(raw_gene_aliases_dict, all_identifiers)
    allele_aliases = combine_species_aliases(raw_allele_aliases_dict, all_identifiers)
    haplotypes = combine_species_aliases(raw_haplotypes_dict, all_identifiers)
    serotypes = combine_species_aliases(raw_serotypes_dict, all_identifiers)
    heterodimers = combine_species_aliases(raw_heterodimers_dict, all_identifiers)
    supertypes = combine_species_aliases(raw_supertypes_dict, all_identifiers)
    known_alleles = combine_species_aliases(
        raw_known_alleles_dict, all_identifiers, value_class=NormalizingSet
    )

    return Species(
        name=latin_name,
        common_name=shortest_common_name,
        mhc_prefix=prefix,
        old_mhc_prefix=old_mhc_prefix,
        gene_names=gene_names,
        gene_name_to_mhc_class=gene_name_to_mhc_class,
        class2_loci=class2_loci,
        class2_locus_to_gene_names=class2_locus_to_gene_names,
        class2_gene_name_to_chain_type=class2_gene_name_to_chain_type,
        gene_aliases=gene_aliases,
        allele_aliases=allele_aliases,
        known_alleles=known_alleles,
        haplotypes=haplotypes,
        serotypes=serotypes,
        heterodimers=heterodimers,
        supertypes=supertypes,
        parent_species=parent_species,
        other_mhc_prefixes=other_mhc_prefixes,
        other_common_names=[name for name in common_names if name != shortest_common_name],
        raw_string=latin_name,
    )


def combine_species_aliases(
    species_dict, species_names, dictionary_class=NormalizingDictionary, value_class=None
):
    if value_class:
        result = dictionary_class(default_value_fn=value_class)
    else:
        result = dictionary_class()
    for species_name in species_names:
        value_for_species = species_dict.get(species_name)
        if value_for_species:
            if type(value_for_species) not in (dict, NormalizingDictionary):
                raise TypeError(f"Expected sub-dictionaries but got {type(value_for_species)}")
            for key, value in value_for_species.items():
                old_value = None
                if key in result:
                    old_value = result[key]
                elif value_class is not None:
                    old_value = value_class()

                if old_value is not None:
                    t = type(old_value)
                    if t in {set, dict, NormalizingSet, NormalizingDictionary}:
                        combined = old_value.copy()
                        combined.update(value)
                        value = combined
                    elif t in {list, tuple}:
                        value = old_value + t(value)
                result[key] = value
    return result


def create_species_lookup_dictionaries():
    gene_name_to_species_objects = NormalizingDictionary(default_value_fn=set)
    # Canonical index: latin name → species (never ambiguous)
    latin_name_to_species = NormalizingDictionary()
    # Legacy index kept for compatibility
    species_name_to_species = NormalizingDictionary()

    # latin name, common names, or MHC prefixes all mapping to multiple
    # species objects
    alias_to_species_objects = NormalizingDictionary(default_value_fn=set)

    for latin_name in raw_species_dict:
        species = create_species_for_latin_name(latin_name)
        latin_name_to_species[latin_name] = species
        species_name_to_species[latin_name] = species
        for s in species.all_identifiers:
            alias_to_species_objects[s].add(species)
        for gene_name in species.gene_names_and_aliases:
            gene_name_to_species_objects[gene_name].add(species)
    return (
        latin_name_to_species,
        species_name_to_species,
        alias_to_species_objects,
        gene_name_to_species_objects,
    )


(
    latin_name_to_species_object,
    species_name_to_species_object,
    alias_to_species_objects,
    gene_name_to_species_objects,
) = create_species_lookup_dictionaries()


def find_matching_species_objects(name):
    """
    Returns list of Species
    """
    if type(name) is Species:
        return [name]
    if name is None:
        return []
    return list(alias_to_species_objects.get(name, []))


def create_species_sort_key(query_string):
    """
    If we get multiple Species matching a query then
    use this key function to sort possible matches
    and prefer ones that match the query best.
    """

    def sort_key(species):
        same_prefix = species.prefix == query_string
        similar_prefix = species.prefix.lower().strip() == query_string.lower().strip()
        num_genes = species.num_genes
        return (same_prefix, similar_prefix, num_genes)

    return sort_key


def infer_species_from_prefix(name, _allow_mhc_strip=True):
    """
    Trying to parse prefixes of alleles such as:
        HLA-A
    but also ones with dashes in the species prefix:
        H-2-K
    and also those lacking any dashes such as:
        H2K

     ...we also need to consider that alleles, haplotypes, etc may come
     immediately after the gene:
        H2Kk
        HLA-A0201

    Returns the Species and the original string that matched it or None.
    """
    # Try parsing a few different substrings to get the species,
    # and then use the species gene list to determine what the gene is in this string
    candidate_species_substrings = {name}

    if "-" in name or "." in name:
        # if name is "H-2-K" or "RT1.A" then try parsing "H", "H-2", "RT1" as
        # species prefixes. We treat both "-" and "." as valid separators.
        parts_split_by_separator = re.split(r"[-.]", name)
        candidate_species_substrings.add(parts_split_by_separator[0])
        if len(parts_split_by_separator) > 1:
            # Also try first two parts joined by hyphen (e.g., "H-2" from "H-2-K")
            candidate_species_substrings.add(
                parts_split_by_separator[0] + "-" + parts_split_by_separator[1]
            )

    for num_chars in [None, 4, 3, 2]:
        for candidate in candidate_species_substrings:
            if num_chars:
                original_prefix = candidate[:num_chars]
            else:
                original_prefix = candidate
            species_objects = find_matching_species_objects(original_prefix)
            if species_objects:
                if len(species_objects) == 1:
                    return species_objects[0], original_prefix
                # Multiple species match this prefix. Try disambiguation:
                # 1. If exactly one species owns this as its primary prefix,
                #    use it (covers parent/child cases like SLA, BoLA, RT1).
                prefix_owners = [
                    sp
                    for sp in species_objects
                    if sp.prefix.lower().strip() == original_prefix.lower().strip()
                ]
                if len(prefix_owners) == 1:
                    return prefix_owners[0], original_prefix
                # 2. Try gene-context disambiguation from the remaining string.
                remaining = name[len(original_prefix) :]
                remaining = remaining.lstrip("-. ")
                before_star = remaining.split("*")[0] if remaining else ""
                gene_candidates = [before_star] if before_star else []
                parts = re.split(r"[-.]", before_star)
                for i in range(len(parts) - 1, 0, -1):
                    gene_candidates.append("-".join(parts[:i]))
                for gene_token in gene_candidates:
                    compatible = [
                        sp
                        for sp in species_objects
                        if sp.find_matching_gene_name(gene_token) is not None
                    ]
                    if len(compatible) == 1:
                        return compatible[0], original_prefix
                # Could not disambiguate — skip this candidate length and
                # try shorter prefixes or other strategies.
                continue

    # Strip a leading "Mhc" prefix commonly seen in bird MHC literature
    # (e.g., "MhcTyal-DAB1*01:01" → "Tyal-DAB1*01:01"). Only attempted
    # once (_allow_mhc_strip flag prevents recursive stripping) and only
    # after normal prefix matching has failed, so a direct prefix match
    # is always preferred.
    if _allow_mhc_strip and len(name) > 3 and name[:3].lower() == "mhc":
        stripped = name[3:]
        result = infer_species_from_prefix(stripped, _allow_mhc_strip=False)
        if result is not None:
            species_object, inner_prefix = result
            # Return the full original prefix including "Mhc" so the caller
            # can correctly compute the remaining string.
            return species_object, name[: 3 + len(inner_prefix)]

    # if all else fails, look for a distinctive gene name which is unique
    # to one species
    if "*" in name:
        parts = name.split("*")
        prefix = parts[0]
        species_objects = Species.get_species_with_gene_name(prefix)
        if len(species_objects) == 1:
            species = species_objects[0]
            # returning an empty string as the second result since no
            # part of the original string actually matched a species name
            # or alias but we inferred it from a gene name
            return species, ""
    return None



import pytest

from mhcgnomes import Allele, Gene, Species, parse

from .common import eq_

CETACEAN_SPECIES = [
    ("Bamu", "Balaenoptera musculus"),
    ("Baac", "Balaenoptera acutorostrata"),
    ("Baed", "Balaenoptera edeni"),
    ("Cehe", "Cephalorhynchus hectori"),
    ("Dede", "Delphinus delphis"),
    ("Euau", "Eubalaena australis"),
    ("Glme", "Globicephala melas"),
    ("Grgr", "Grampus griseus"),
    ("Kobr", "Kogia breviceps"),
    ("Laob", "Lagenorhynchus obscurus"),
    ("Mede", "Mesoplodon densirostris"),
    ("Meeu", "Mesoplodon europaeus"),
    ("Megr", "Mesoplodon grayi"),
    ("Meno", "Megaptera novaeangliae"),
    ("Stbr", "Steno bredanensis"),
    ("Stco", "Stenella coeruleoalba"),
    ("Tutr", "Tursiops truncatus"),
    ("Zica", "Ziphius cavirostris"),
]

CETACEAN_REPRESENTATIVE_ALLELES = [
    ("Bamu-DQA*001:01:01", "Bamu", "DQA", ("001", "01", "01")),
    ("Baac-S*001:01", "Baac", "S", ("001", "01")),
    ("Baed-DQB*001:01", "Baed", "DQB", ("001", "01")),
    ("Cehe-DRA*001:01", "Cehe", "DRA", ("001", "01")),
    ("Dede-N*001:01", "Dede", "N", ("001", "01")),
    ("Euau-DQA*001:01:01", "Euau", "DQA", ("001", "01", "01")),
    ("Glme-DQB*001:01", "Glme", "DQB", ("001", "01")),
    ("Grgr-N*001:01", "Grgr", "N", ("001", "01")),
    ("Kobr-DRA*001:01", "Kobr", "DRA", ("001", "01")),
    ("Laob-N*001:01", "Laob", "N", ("001", "01")),
    ("Mede-DQA*001:01:01", "Mede", "DQA", ("001", "01", "01")),
    ("Meeu-DQB*001:01", "Meeu", "DQB", ("001", "01")),
    ("Megr-DRA*001:01", "Megr", "DRA", ("001", "01")),
    ("Meno-DQB*001:01", "Meno", "DQB", ("001", "01")),
    ("Stbr-N*001:01", "Stbr", "N", ("001", "01")),
    ("Stco-DRA*001:01", "Stco", "DRA", ("001", "01")),
    ("Tutr-N*001:01", "Tutr", "N", ("001", "01")),
    ("Zica-DRA*001:01", "Zica", "DRA", ("001", "01")),
]

CETACEAN_GENE_EXAMPLES = [
    ("Bamu-DQA*001:01:01", "Bamu", "DQA", "IIa", ("001", "01", "01"), "alpha"),
    ("Bamu-DQB*001:01", "Bamu", "DQB", "IIa", ("001", "01"), "beta"),
    ("Bamu-DRA*001:01", "Bamu", "DRA", "IIa", ("001", "01"), "alpha"),
    ("Bamu-DRB1*001:01", "Bamu", "DRB1", "IIa", ("001", "01"), "beta"),
    ("Bamu-N*001:01", "Bamu", "N", "I", ("001", "01"), None),
    ("Baac-S*001:01", "Baac", "S", "I", ("001", "01"), None),
]


@pytest.mark.parametrize("species_prefix,species_name", CETACEAN_SPECIES)
def test_cetacean_species_registered(species_prefix, species_name):
    species = Species.get(species_prefix)
    assert species is not None
    eq_(species.name, species_name)
    eq_(species.historic_mhc_prefix, "CELA")


@pytest.mark.parametrize(
    "allele_name,species_prefix,gene_name,mhc_class,fields,chain_type",
    CETACEAN_GENE_EXAMPLES,
)
def test_cetacean_gene_metadata(
    allele_name, species_prefix, gene_name, mhc_class, fields, chain_type
):
    gene = Gene.get(species_prefix, gene_name)
    assert gene is not None
    eq_(gene.species.prefix, species_prefix)
    eq_(gene.name, gene_name)
    eq_(gene.mhc_class, mhc_class)

    parsed_gene = parse(f"{species_prefix}-{gene_name}")
    eq_(parsed_gene, gene)

    allele = parse(allele_name)
    eq_(type(allele), Allele)
    eq_(allele.species.prefix, species_prefix)
    eq_(allele.gene.name, gene_name)
    eq_(allele.allele_fields, fields)

    if chain_type == "alpha":
        assert allele.is_class2_alpha
    elif chain_type == "beta":
        assert allele.is_class2_beta
    else:
        assert allele.is_class1


@pytest.mark.parametrize(
    "allele_name,species_prefix,gene_name,fields",
    CETACEAN_REPRESENTATIVE_ALLELES,
)
def test_parse_representative_cetacean_alleles(allele_name, species_prefix, gene_name, fields):
    allele = parse(allele_name)
    eq_(type(allele), Allele)
    eq_(allele.species.prefix, species_prefix)
    eq_(allele.gene.name, gene_name)
    eq_(allele.allele_fields, fields)


def test_cela_old_prefix_round_trip():
    species = Species.get("CELA")
    assert species is not None
    eq_(species.name, "Cetacea sp.")

    allele = parse("Tutr-DQA*001:01:01")
    eq_(allele.to_string(use_old_species_prefix=True), "CELA-DQA*001:01:01")


def test_parse_historic_cela_input_uses_generic_species():
    gene = Gene.get("CELA", "DQA")
    assert gene is not None
    eq_(gene.species.prefix, "CELA")
    eq_(gene.mhc_class, "IIa")

    allele = parse("CELA-DQA*001:01:01")
    eq_(type(allele), Allele)
    eq_(allele.species.prefix, "CELA")
    eq_(allele.gene.name, "DQA")
    eq_(allele.allele_fields, ("001", "01", "01"))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from . import function_api


def dataframe_from_parsed_objects(parsed_objects):
    """
    Create a pandas DataFrame from a list of parsed MHC objects.

    Each parsed object's `to_record()` method is called to generate a
    dictionary, which is then converted to a DataFrame row.

    Parameters
    ----------
    parsed_objects : list of Result
        List of parsed MHC objects (Allele, Gene, Species, etc.).

    Returns
    -------
    pandas.DataFrame
        DataFrame with columns corresponding to the object fields.

    Examples
    --------
    >>> from mhcgnomes import parse
    >>> objects = [parse("HLA-A*02:01"), parse("HLA-B*07:02")]
    >>> df = dataframe_from_parsed_objects(objects)
    >>> df["gene"].tolist()
    ['HLA-A', 'HLA-B']
    """
    records = [obj.to_record() for obj in parsed_objects]
    return pd.DataFrame.from_records(records)


def dataframe_from_string_list(names):
    """
    Parse a list of MHC allele strings and return a DataFrame.

    This is a convenience function that parses each string and creates
    a DataFrame from the results.

    Parameters
    ----------
    names : list of str
        List of MHC allele/gene/species strings to parse.

    Returns
    -------
    pandas.DataFrame
        DataFrame with parsed information for each input string.

    Examples
    --------
    >>> names = ["HLA-A*02:01", "HLA-B*07:02", "HLA-C*07:01"]
    >>> df = dataframe_from_string_list(names)
    >>> df["allele"].tolist()
    ['HLA-A*02:01', 'HLA-B*07:02', 'HLA-C*07:01']
    """
    parsed_objects = [function_api.parse(name) for name in names]
    return dataframe_from_parsed_objects(parsed_objects)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .common import normalize_string


class NormalizingSet:
    """
    Like a regular set but all keys get normalized by a user
    provided function.
    """

    def __init__(self, *items, normalize_fn=normalize_string):
        self.item_to_original = {}
        self.items = set()
        self.normalize_fn = normalize_fn
        self._frozen = False
        self.update(items)

    def copy(self):
        return NormalizingSet(*list(self), normalize_fn=self.normalize_fn)

    def freeze(self):
        self._frozen = True
        return self

    @property
    def is_frozen(self):
        return self._frozen

    def _check_mutable(self):
        if self._frozen:
            raise TypeError("NormalizingSet is frozen")

    def __contains__(self, item):
        item = self.normalize_fn(item)
        return item in self.items

    def __eq__(self, other):
        if type(other) is not NormalizingSet:
            return False
        if len(self) != len(other):
            return False
        return self.items == other.items

    def get_original(self, item):
        normalized = self.normalize_fn(item)
        return self.item_to_original.get(normalized)

    def add(self, extra_item):
        self._check_mutable()
        normalized = self.normalize_fn(extra_item)
        self.item_to_original[normalized] = extra_item
        self.items.add(normalized)

    def update(self, extra_items):
        self._check_mutable()
        for original in extra_items:
            self.add(original)

    def __iter__(self):
        for normalized in self.items:
            yield self.item_to_original[normalized]

    def __len__(self):
        return len(self.items)



"""
Tests for the mhc1/mhc2 gene prefix stripping in Species.find_matching_gene_name().

ZFIN and some other fish databases use lowercase concatenated gene names like
"mhc1uba" (class I gene UBA) and "mhc2daa" (class II gene DAA). The parser
strips mhc1/mhc2 prefixes generically rather than requiring per-gene aliases.
"""

from mhcgnomes import Gene, parse

from .common import eq_

# ---------------------------------------------------------------------------
# Positive tests: mhc1/mhc2 prefix stripping resolves known genes
# ---------------------------------------------------------------------------


class TestMhc1Mhc2PrefixStrippingPositive:
    """Gene tokens starting with mhc1/mhc2 resolve to the stripped gene name."""

    def test_dare_mhc1uba(self):
        eq_(parse("Dare-mhc1uba", raise_on_error=True), Gene.get("Dare", "UBA"))

    def test_dare_mhc1uma(self):
        eq_(parse("Dare-mhc1uma", raise_on_error=True), Gene.get("Dare", "UMA"))

    def test_dare_mhc2daa(self):
        eq_(parse("Dare-mhc2daa", raise_on_error=True), Gene.get("Dare", "DAA"))

    def test_dare_mhc2dab(self):
        eq_(parse("Dare-mhc2dab", raise_on_error=True), Gene.get("Dare", "DAB"))

    def test_case_insensitive_MHC1UBA(self):
        """Uppercase MHC1UBA should also strip."""
        eq_(parse("Dare-MHC1UBA", raise_on_error=True), Gene.get("Dare", "UBA"))

    def test_case_insensitive_MHC2DAB(self):
        eq_(parse("Dare-MHC2DAB", raise_on_error=True), Gene.get("Dare", "DAB"))

    def test_works_for_other_fish_with_uba(self):
        """mhc1 stripping should work for any fish species, not just zebrafish."""
        eq_(parse("Saal-mhc1uba", raise_on_error=True), Gene.get("Saal", "UBA"))

    def test_works_for_salmon_mhc2dab(self):
        eq_(parse("Sasa-mhc2dab", raise_on_error=True), Gene.get("Sasa", "DAB"))

    def test_works_for_trout_mhc1uba(self):
        eq_(parse("Onmy-mhc1uba", raise_on_error=True), Gene.get("Onmy", "UBA"))

    def test_allele_with_mhc1_prefix(self):
        """mhc1/mhc2 stripping should work in allele context too."""
        result = parse("Dare-mhc1uba*01", raise_on_error=True)
        eq_(result.gene.name, "UBA")

    def test_allele_with_mhc2_prefix(self):
        result = parse("Dare-mhc2daa*01:01", raise_on_error=True)
        eq_(result.gene.name, "DAA")


# ---------------------------------------------------------------------------
# Negative tests: mhc1/mhc2 stripping does not create false positives
# ---------------------------------------------------------------------------


class TestMhc1Mhc2PrefixStrippingNegative:
    """Stripping must not match genes that don't exist in the species."""

    def test_mhc1_with_nonexistent_gene_returns_none(self):
        """mhc1zzz should not parse — ZZZ is not a zebrafish gene."""
        assert parse("Dare-mhc1zzz", raise_on_error=False) is None

    def test_mhc2_with_nonexistent_gene_returns_none(self):
        assert parse("Dare-mhc2zzz", raise_on_error=False) is None

    def test_mhc1_alone_parses_as_mhc_class(self):
        """Bare 'mhc1' parses as MhcClass(I), not a Gene."""
        from mhcgnomes import MhcClass

        result = parse("Dare-mhc1", raise_on_error=False)
        assert result is None or isinstance(result, MhcClass)

    def test_mhc2_alone_parses_as_mhc_class(self):
        from mhcgnomes import MhcClass

        result = parse("Dare-mhc2", raise_on_error=False)
        assert result is None or isinstance(result, MhcClass)

    def test_mhc3_prefix_not_stripped(self):
        """Only mhc1 and mhc2 should be stripped, not mhc3 or other numbers."""
        assert parse("Dare-mhc3uba", raise_on_error=False) is None

    def test_mhc_without_number_not_stripped_at_gene_level(self):
        """Plain 'mhc' prefix on a gene token should not be stripped.
        (This is handled at the species level, not gene level.)"""
        # mhcuba is not a valid gene — "mhc" without 1/2 is not a gene prefix
        assert parse("Dare-mhcuba", raise_on_error=False) is None

    def test_does_not_strip_for_species_without_gene(self):
        """mhc1dqa on a species that doesn't have DQA should fail."""
        # Zebrafish has UBA, UMA, DAA, DAB — but not DQA
        assert parse("Dare-mhc1dqa", raise_on_error=False) is None

    def test_does_not_strip_when_gene_exists_directly(self):
        """If a gene alias 'mhc1uba' were also defined directly, the
        direct match should win. Here we just verify direct genes still
        work when they don't start with mhc1/mhc2."""
        # UBA matches directly, no stripping needed
        eq_(parse("Dare-UBA", raise_on_error=True), Gene.get("Dare", "UBA"))



"""
Tests for the Mhc-prefix stripping logic in infer_species_from_prefix.

The parser strips a leading "Mhc" (case-insensitive) from allele strings
when normal prefix matching fails, so that literature-style names like
"MhcTyal-DAB1*01:01" resolve to the correct species and gene. This
stripping is a last-resort fallback — a direct prefix match always wins.
"""

from mhcgnomes import Allele, Gene, parse
from mhcgnomes.species import infer_species_from_prefix

from .common import eq_

# ---------------------------------------------------------------------------
# 1. Basic Mhc stripping for barn owl (the motivating case)
# ---------------------------------------------------------------------------


def test_mhc_strip_tyal_species_inferred():
    result = infer_species_from_prefix("MhcTyal-DAB1*01:01")
    assert result is not None
    species, prefix = result
    eq_(species.prefix, "TytoAlba")
    eq_(prefix, "MhcTyal")


def test_mhc_strip_tyal_gene():
    eq_(parse("MhcTyal-UA", raise_on_error=True), Gene.get("TytoAlba", "UA"))


def test_mhc_strip_tyal_gene_dab1():
    eq_(parse("MhcTyal-DAB1", raise_on_error=True), Gene.get("TytoAlba", "DAB1"))


def test_mhc_strip_tyal_gene_dab2():
    eq_(parse("MhcTyal-DAB2", raise_on_error=True), Gene.get("TytoAlba", "DAB2"))


def test_mhc_strip_tyal_allele_ua():
    expected = Allele.get("TytoAlba", "UA", "01", "01")
    assert expected is not None
    eq_(parse("MhcTyal-UA*01:01", raise_on_error=True), expected)


def test_mhc_strip_tyal_allele_dab1():
    expected = Allele.get("TytoAlba", "DAB1", "01", "01")
    assert expected is not None
    eq_(parse("MhcTyal-DAB1*01:01", raise_on_error=True), expected)


# ---------------------------------------------------------------------------
# 2. Works for other bird species (not just barn owl)
# ---------------------------------------------------------------------------


def test_mhc_strip_acar_gene():
    """Great reed warbler — MhcAcar-UA should parse."""
    eq_(parse("MhcAcar-UA", raise_on_error=True), Gene.get("AcroArun", "UA"))


def test_mhc_strip_acar_allele():
    expected = Allele.get("AcroArun", "UA", "01", "01")
    assert expected is not None
    eq_(parse("MhcAcar-UA*01:01", raise_on_error=True), expected)


def test_mhc_strip_fuat_gene():
    """Eurasian coot — MhcFuat-DAB should parse."""
    eq_(parse("MhcFuat-DAB", raise_on_error=True), Gene.get("Fuat", "DAB"))


def test_mhc_strip_spma_gene():
    """Magellanic penguin — MhcSpma-DRB1 should parse."""
    eq_(parse("MhcSpma-DRB1", raise_on_error=True), Gene.get("Spma", "DRB1"))


def test_mhc_strip_coja_gene():
    """Japanese quail — MhcCoja-DAB1 should parse."""
    eq_(parse("MhcCoja-DAB1", raise_on_error=True), Gene.get("Coja", "DAB1"))


def test_mhc_strip_sthi_gene():
    """Common tern — MhcSthi-DAB should parse."""
    eq_(parse("MhcSthi-DAB", raise_on_error=True), Gene.get("Sthi", "DAB"))


# ---------------------------------------------------------------------------
# 3. Works for non-bird species too
# ---------------------------------------------------------------------------


def test_mhc_strip_paol_gene():
    """Olive flounder — MhcPaol-DAB should parse."""
    eq_(parse("MhcPaol-DAB", raise_on_error=True), Gene.get("Paol", "DAB"))


def test_mhc_strip_dare_gene():
    """Zebrafish — MhcDare-UBA should parse."""
    eq_(parse("MhcDare-UBA", raise_on_error=True), Gene.get("Dare", "UBA"))


def test_mhc_strip_xela_gene():
    """African clawed frog — MhcXela-UAA should parse."""
    eq_(parse("MhcXela-UAA", raise_on_error=True), Gene.get("Xela", "UAA"))


def test_mhc_strip_sasa_allele():
    """Atlantic salmon — MhcSasa-DAB*05:01 should parse."""
    expected = Allele.get("Sasa", "DAB", "05", "01")
    assert expected is not None
    eq_(parse("MhcSasa-DAB*05:01", raise_on_error=True), expected)


# ---------------------------------------------------------------------------
# 4. Case insensitivity of the Mhc prefix
# ---------------------------------------------------------------------------


def test_mhc_strip_lowercase():
    eq_(parse("mhcTyal-UA", raise_on_error=True), Gene.get("TytoAlba", "UA"))


def test_mhc_strip_uppercase():
    eq_(parse("MHCTyal-UA", raise_on_error=True), Gene.get("TytoAlba", "UA"))


def test_mhc_strip_mixed_case():
    eq_(parse("mHcTyal-UA", raise_on_error=True), Gene.get("TytoAlba", "UA"))


# ---------------------------------------------------------------------------
# 5. Direct prefix match always wins over Mhc stripping
# ---------------------------------------------------------------------------


def test_hla_not_stripped():
    """HLA starts with H, not Mhc — should parse normally."""
    result = infer_species_from_prefix("HLA-A*02:01")
    assert result is not None
    species, _prefix = result
    eq_(species.prefix, "HLA")


def test_direct_prefix_preferred_over_strip():
    """
    Known prefixes that happen to start with M are not mangled.
    """
    # Maar = Macaca arctoides (stump-tailed macaque)
    result = infer_species_from_prefix("Maar-A1")
    assert result is not None
    species, _prefix = result
    eq_(species.prefix, "Maar")


def test_mhcmafa_direct_prefix_wins():
    """
    MhcMafa is a registered alternative prefix for Macaca fascicularis.
    The direct prefix match must win — stripping 'Mhc' and matching 'Mafa'
    would also work, but the direct match is preferred and returns the
    correct full prefix length.
    """
    result = infer_species_from_prefix("MhcMafa-I*01:01")
    assert result is not None
    species, prefix = result
    eq_(species.prefix, "Mafa")
    # The full "MhcMafa" should be consumed as the prefix, not just "Mafa"
    eq_(prefix, "MhcMafa")


def test_gaga_direct_match_not_stripped():
    """Gaga (chicken) should match directly, not via Mhc stripping."""
    result = infer_species_from_prefix("Gaga-BF1")
    assert result is not None
    species, prefix = result
    eq_(species.prefix, "Gaga")
    eq_(prefix, "Gaga")


# ---------------------------------------------------------------------------
# 6. Stripping does NOT happen when it would produce garbage
# ---------------------------------------------------------------------------


def test_mhc_alone_returns_none():
    """Bare 'Mhc' with nothing after it should not match."""
    result = infer_species_from_prefix("Mhc")
    assert result is None


def test_mhc_with_unknown_suffix_returns_none():
    """Mhc followed by nonsense should not match any species."""
    result = infer_species_from_prefix("MhcZZZZ-FOO")
    assert result is None


def test_mhc_strip_does_not_double_strip():
    """
    If stripping 'Mhc' once leaves another 'Mhc' prefix, it should NOT
    strip again. E.g., 'MhcMhcTyal' should not resolve to Tyal.
    """
    result = infer_species_from_prefix("MhcMhcTyal-UA")
    assert result is None


# ---------------------------------------------------------------------------
# 7. The returned prefix length is correct for remaining-string computation
# ---------------------------------------------------------------------------


def test_mhc_strip_remaining_string_gene():
    """After stripping MhcTyal, remaining string should be '-UA'."""
    result = infer_species_from_prefix("MhcTyal-UA")
    assert result is not None
    _species, prefix = result
    remaining = "MhcTyal-UA"[len(prefix) :]
    eq_(remaining, "-UA")


def test_mhc_strip_remaining_string_allele():
    """After stripping MhcAcar, remaining string should be '-UA*01:01'."""
    result = infer_species_from_prefix("MhcAcar-UA*01:01")
    assert result is not None
    _species, prefix = result
    remaining = "MhcAcar-UA*01:01"[len(prefix) :]
    eq_(remaining, "-UA*01:01")


def test_mhc_strip_prefix_length_equals_mhc_plus_species():
    """The returned prefix should be exactly 'Mhc' + species prefix."""
    result = infer_species_from_prefix("MhcFuat-DAB*199")
    assert result is not None
    _species, prefix = result
    eq_(prefix, "MhcFuat")
    eq_(len(prefix), 3 + len("Fuat"))


# ---------------------------------------------------------------------------
# 8. Integration: full parse round-trips with Mhc prefix
# ---------------------------------------------------------------------------


def test_mhc_strip_round_trip_gene_to_string():
    """Parsed MhcTyal-UA should normalize to 'Tyal-UA' in to_string()."""
    result = parse("MhcTyal-UA", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.prefix, "TytoAlba")
    eq_(result.name, "UA")


def test_mhc_strip_round_trip_allele_to_string():
    """Parsed MhcAcar-UA*01:01 should produce a valid Allele with Acar species."""
    result = parse("MhcAcar-UA*01:01", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "AcroArun")
    eq_(result.gene.name, "UA")


def test_mhc_strip_round_trip_allele_fields():
    """Allele fields should be correctly parsed after Mhc stripping."""
    result = parse("MhcTyal-DAB1*03:02", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "TytoAlba")
    eq_(result.gene.name, "DAB1")
    eq_(result.allele_fields, ("03", "02"))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from dataclasses import dataclass
from typing import Union


@dataclass(eq=False, repr=False, frozen=True, init=False)
class Result:
    """
    Base class for all parsed objects in mhcgnomes.
    """

    raw_string: Union[str, None] = None

    def __init__(self, raw_string=None):
        object.__setattr__(self, "raw_string", raw_string)

    @staticmethod
    def _set_field(instance, field_name, field_value):
        object.__setattr__(instance, field_name, field_value)

    @classmethod
    def init_field_names(cls):
        """
        Which fields are required to create a new instance of this object?

        By default extract everything except "self", "*args", and "**kwargs".
        """
        sig = inspect.signature(cls.__init__)
        params = sig.parameters
        init_arg_names = tuple(
            [
                k
                for k, p in params.items()
                if k != "self" and p.kind not in {p.VAR_KEYWORD, p.VAR_POSITIONAL}
            ]
        )
        return init_arg_names

    @classmethod
    def init_field_names_without_raw_string(cls):
        """
        Remove 'raw_string' from field names list to avoid
        e.g. having objects with identical

        """
        return tuple([x for x in cls.init_field_names() if x != "raw_string"])

    @classmethod
    def str_field_names(cls):
        """
        Which fields are includes in __str__ string.
        """
        return cls.init_field_names_without_raw_string()

    @classmethod
    def repr_field_names(cls):
        """
        Which fields are includes in __repr__ string.
        """
        return cls.str_field_names()

    @classmethod
    def tuple_field_names(cls):
        """
        Which fields are includes in the dict and tuple representations
        of this object.
        """
        return cls.init_field_names()

    @classmethod
    def eq_field_names(cls):
        """
        Which fields are includes in equality comparison.
        """
        return cls.init_field_names_without_raw_string()

    @classmethod
    def hash_field_names(cls):
        """
        Which fields are includes in the dict and tuple representations
        of this object.
        """
        return cls.eq_field_names()

    def _field_name_value_pairs(self, names):
        results = []
        for field_name in names:
            field_value = getattr(self, field_name)
            results.append((field_name, field_value))
        return results

    def _field_name_string_pairs(self, names):
        results = []
        for k, v in self._field_name_value_pairs(names):
            if isinstance(v, str):
                results.append((k, f"'{v}'"))
            else:
                results.append((k, f"{v}"))
        return results

    def __str__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join(
                [
                    f"{k}={v_str}"
                    for (k, v_str) in self._field_name_string_pairs(self.str_field_names())
                ]
            ),
        )

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join(
                [
                    f"{k}={v_str}"
                    for (k, v_str) in self._field_name_string_pairs(self.repr_field_names())
                ]
            ),
        )

    def to_string(self, include_species=True, use_old_species_prefix=False):
        raise NotImplementedError(
            f"{self.__class__.__name__} requires implementation of to_string() method"
        )

    def compact_string(self, include_species=False, use_old_species_prefix=False):
        """
        Compact representation, defaults to omitting species
        """
        return self.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        return all(getattr(self, field) == getattr(other, field) for field in self.eq_field_names())

    def __hash__(self):
        total = 0
        for field in self.hash_field_names():
            total += hash(getattr(self, field))
        return total

    def to_record(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} requires implementation of to_record() method"
        )

    def to_tuple(self):
        keys = self.tuple_field_names()
        values = [getattr(self, k) for k in keys]
        return tuple(values)

    @classmethod
    def from_tuple(cls, t):
        keys = cls.tuple_field_names()
        assert len(keys) == len(t)
        d = dict(zip(keys, t))
        return cls.from_dict(d)

    def to_dict(self):
        return dict(zip(self.tuple_field_names(), self.to_tuple()))

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    def copy(self, **kwargs):
        """
        Make a copy of this object and update any specified fields.
        """
        field_dict = self.to_dict()
        field_dict.update(kwargs)
        return self.__class__.from_dict(field_dict)

    def __lt__(self, other):
        self_key = (self.__class__.__name__, *self.to_tuple())
        other_key = (other.__class__.__name__, *other.to_tuple())
        return self_key < other_key

    ############################################################################
    #
    #  Default properties shared across all result objects
    #
    ############################################################################

    @property
    def annotation_null(self):
        return False

    @property
    def annotation_cystosolic(self):
        return False

    @property
    def annotation_secreted(self):
        return False

    @property
    def annotation_questionable(self):
        return False

    @property
    def annotation_low_expression(self):
        return False

    @property
    def annotation_aberrant_expression(self):
        return False

    @property
    def annotation_group(self):
        # designates a group of genomic sequence alleles
        # with identical peptide binding region
        return False

    @property
    def annotation_pseudogene(self):
        # designates a group of genomic sequence alleles
        # with identical peptide binding region
        return False

    @property
    def annotation_splice_variant(self):
        # designates a group of genomic sequence alleles
        # with identical peptide binding region
        return False

    @property
    def is_class1(self):
        return False

    @property
    def is_class2(self):
        return False

    @property
    def is_class2_alpha(self):
        return False

    @property
    def is_class2_beta(self):
        return False

    @property
    def has_species(self):
        return False

    @property
    def has_mhc_class(self):
        return False



from mhcgnomes import Haplotype, parse

from .common import eq_


def test_parse_haplotype_ELA_A1():
    result = parse("ELA-A1")
    assert result is not None
    assert type(result) is Haplotype
    eq_(result.name, "A1")
    assert result.species.is_horse


def test_parse_haplotype_ELA_A1_lowercase():
    result = parse("ela-a1")
    assert result is not None
    assert type(result) is Haplotype
    eq_(result.name, "A1")
    assert result.species.is_horse



"""
Tests for handling nonsense and malformed inputs.

These tests verify that mhcgnomes gracefully handles invalid inputs
without crashing, and documents the parser's behavior with edge cases.

The parser is intentionally lenient to handle the many variations
found in real-world data (different separators, case variations, etc.).
"""

import pytest

from mhcgnomes import parse
from mhcgnomes.allele import Allele
from mhcgnomes.errors import ParseError
from mhcgnomes.gene import Gene
from mhcgnomes.species import Species


class TestCompleteNonsense:
    """Tests for completely invalid inputs that should be rejected."""

    @pytest.mark.parametrize(
        "inp",
        [
            # Random strings
            "asdfghjkl",
            "hello world",
            "12345",
            "!!!@@@###",
            # Empty/whitespace
            "",
            "   ",
            # Other genes
            "BRCA1",
            "TP53",
            "KIR2DL1",
            "CD4",
            "CD8",
            # Injection attempts (should fail, not crash)
            "'; DROP TABLE alleles;--",
            "<script>alert('xss')</script>",
            # Unicode
            "🧬",
            "HLA-Ä1",
        ],
    )
    def test_nonsense_raises_parse_error(self, inp):
        """Complete nonsense should raise ParseError, not crash."""
        with pytest.raises(ParseError):
            parse(inp)

    @pytest.mark.parametrize(
        "inp",
        [
            "null",
            "None",
            "undefined",
            "NaN",
        ],
    )
    def test_null_like_strings_rejected(self, inp):
        """Strings that look like null values should be rejected."""
        with pytest.raises(ParseError):
            parse(inp)


class TestPartialInputs:
    """Tests for partial/incomplete inputs."""

    def test_hla_prefix_alone_returns_species(self):
        """Just 'HLA' should return the Species object."""
        result = parse("HLA")
        assert isinstance(result, Species)
        assert result.prefix == "HLA"

    def test_hla_with_dash_returns_species(self):
        """'HLA-' (with trailing dash) returns Species."""
        result = parse("HLA-")
        assert isinstance(result, Species)

    def test_gene_without_allele_returns_gene(self):
        """'HLA-A*' (gene without allele fields) returns Gene."""
        result = parse("HLA-A*")
        assert isinstance(result, Gene)
        assert result.name == "A"

    @pytest.mark.parametrize(
        "inp",
        [
            "HLA-X",  # X is not a valid gene
            "HLA-Z99",
            "Z*01:01",  # Z is not a valid gene letter
        ],
    )
    def test_invalid_genes_rejected(self, inp):
        """Invalid gene names should be rejected."""
        with pytest.raises(ParseError):
            parse(inp)


class TestMalformedAlleles:
    """Tests for malformed allele strings."""

    @pytest.mark.parametrize(
        "inp",
        [
            "A*:01",  # Empty first field
            "A*01:",  # Empty second field
            "A*AB:CD",  # Letters instead of numbers
            "A*01.5:01",  # Decimal number
            "A*01:01:01:01:01:01",  # Too many fields (6)
        ],
    )
    def test_clearly_malformed_rejected(self, inp):
        """Clearly malformed allele strings should be rejected."""
        with pytest.raises(ParseError):
            parse(inp)


class TestLenientParsing:
    """
    Tests documenting the parser's lenient behavior.

    The parser is intentionally flexible to handle real-world data variations.
    These tests document what IS accepted (not necessarily what SHOULD be).
    """

    def test_fake_allele_numbers_accepted(self):
        """
        Fake allele numbers are accepted - no database validation.

        This is by design: the parser doesn't validate against the IMGT/HLA
        database. It accepts any syntactically valid allele string.
        """
        result = parse("A*99:99")
        assert isinstance(result, Allele)
        assert result.allele_fields == ("99", "99")

    def test_zero_allele_fields_accepted(self):
        """
        A*00:00 is accepted even though it's not a real allele.

        This might be undesirable but is current behavior.
        """
        result = parse("A*00:00")
        assert isinstance(result, Allele)
        assert result.allele_fields == ("00", "00")

    def test_various_field_lengths_accepted(self):
        """Different allele field lengths are accepted."""
        # Single digit
        r1 = parse("A*1:1")
        assert r1.allele_fields == ("1", "1")

        # Two digit (standard)
        r2 = parse("A*01:01")
        assert r2.allele_fields == ("01", "01")

        # Three digit
        r3 = parse("A*001:001")
        assert r3.allele_fields == ("001", "001")

    def test_underscore_separator_normalized(self):
        """Underscore separators are accepted and normalized to colons."""
        result = parse("A*01_01")
        assert isinstance(result, Allele)
        assert result.to_string() == "HLA-A*01:01"

    def test_flexible_species_separator(self):
        """Parser accepts various separators after species prefix."""
        # Standard
        r1 = parse("HLA-A*01:01")
        # Asterisk instead of dash
        r2 = parse("HLA*A*01:01")
        # Colon instead of dash
        r3 = parse("HLA:A:01:01")

        # All should produce equivalent results
        assert r1.to_string() == r2.to_string() == r3.to_string() == "HLA-A*01:01"

    def test_unknown_annotation_accepted(self):
        """
        Unknown annotation suffixes are accepted.

        Valid annotations: N (null), L (low), S (secreted), Q (questionable), A (aberrant)
        But parser accepts any single letter.
        """
        result = parse("A*01:01X")
        assert isinstance(result, Allele)


class TestCaseInsensitivity:
    """Tests for case-insensitive parsing."""

    @pytest.mark.parametrize(
        "inp",
        [
            "HLA",
            "hla",
            "Hla",
            "hLa",
        ],
    )
    def test_species_case_insensitive(self, inp):
        """Species prefix should be case-insensitive."""
        result = parse(inp)
        assert isinstance(result, Species)

    @pytest.mark.parametrize(
        "inp",
        [
            "HLA-A2",
            "hla-a2",
            "HLA-a2",
            "Hla-A2",
        ],
    )
    def test_serotype_case_insensitive(self, inp):
        """Serotype parsing should be case-insensitive."""
        result = parse(inp)
        assert result.name == "A2"


class TestSpreadsheetNomenclature:
    """Tests for nomenclature variations commonly seen in spreadsheets."""

    @pytest.mark.parametrize(
        "inp,expected_type",
        [
            # Mouse H2 haplotypes - various formats
            ("H2k", "Gene"),
            ("H-2k", "Gene"),
            ("H2-k", "Gene"),
            ("H-2-k", "Gene"),
            ("haplotype H2k", "Haplotype"),
            ("H2k haplotype", "Haplotype"),
            # Human with "serotype" keyword
            ("HLA-A2 serotype", "Serotype"),
            ("serotype A2", "Serotype"),
            # MHC class descriptors
            ("HLA class I", "MhcClass"),
            ("HLA class II", "MhcClass"),
            ("MHC class I", "MhcClass"),
            ("class I HLA", "MhcClass"),
            # Rat RT1
            ("RT1", "Species"),
            ("RT1-A", "Gene"),
            # Multi-field alleles
            ("HLA-A*02:01:01:01", "Allele"),
            ("A*02:01:01:01", "Allele"),
            ("HLA-DRB1*04:01", "Allele"),
            ("DRB1*0401", "Allele"),
            # Alleles with annotations
            ("HLA-A*01:01N", "Allele"),
            ("A*24:02:01:02L", "Allele"),
        ],
    )
    def test_spreadsheet_nomenclature_accepted(self, inp, expected_type):
        """Common spreadsheet nomenclature should parse correctly."""
        result = parse(inp)
        assert type(result).__name__ == expected_type, f"{inp} should parse as {expected_type}"

    @pytest.mark.parametrize(
        "inp",
        [
            "A2 antigen",  # "antigen" suffix not supported
            "HLA-B27 antigen",
        ],
    )
    def test_unsupported_nomenclature_rejected(self, inp):
        """Some nomenclature variations are not supported."""
        with pytest.raises(ParseError):
            parse(inp)

    @pytest.mark.parametrize(
        "inp,expected_name",
        [
            ("B17.1", "B17.1"),
            ("B17.2", "B17.2"),
            ("B17.3", "B17.3"),
            ("HLA-B17.1", "B17.1"),
        ],
    )
    def test_b17_subsplits_with_dots(self, inp, expected_name):
        """B17 subsplits use dot notation and should parse."""
        from mhcgnomes.serotype import Serotype

        result = parse(inp)
        assert isinstance(result, Serotype)
        assert result.name == expected_name


class TestDQDRHeterodimers:
    """
    Tests for HLA-DQ heterodimer dot nomenclature.

    For comprehensive heterodimer tests, see tests/test_heterodimers.py.

    This class only contains basic sanity checks to document that:
    1. DQ2 (without dot) parses as Serotype
    2. DQ2.5 (with dot) should parse as Pair (heterodimer) - currently xfail
    """

    def test_dq2_without_dot_is_serotype(self):
        """DQ2 without dot notation parses as Serotype."""
        from mhcgnomes.serotype import Serotype

        result = parse("DQ2")
        assert isinstance(result, Serotype)
        assert result.name == "DQ2"

    def test_dq2_5_with_dot_is_pair(self):
        """DQ2.5 with dot notation should parse as Pair (heterodimer)."""
        from mhcgnomes.pair import Pair

        result = parse("DQ2.5")
        assert isinstance(result, Pair)


class TestNonHLAGenes:
    """Tests confirming that non-HLA/MHC genes are rejected."""

    @pytest.mark.parametrize(
        "inp",
        [
            "BRCA1",
            "BRCA2",
            "TP53",
            "EGFR",
            "KRAS",
            "insulin",
            "hemoglobin",
            "COVID-19",
            "SARS-CoV-2",
        ],
    )
    def test_non_mhc_genes_rejected(self, inp):
        """Non-MHC gene names should be rejected."""
        with pytest.raises(ParseError):
            parse(inp)



import argparse
import json
import sys
from collections.abc import Sequence
from typing import Optional

from .errors import ParseError
from .function_api import parse

COLUMN_NAMES = (
    "input",
    "type",
    "normalized",
    "compact",
    "species",
    "gene",
    "mhc_class",
    "properties",
)


def _stringify(value):
    if value is None:
        return ""
    if isinstance(value, (list, tuple, set)):
        return ",".join(_stringify(x) for x in value)
    if isinstance(value, dict):
        return ",".join(f"{k}:{_stringify(v)}" for k, v in value.items())
    return str(value)


def _record_to_properties(record) -> str:
    if not record:
        return ""
    if not isinstance(record, dict):
        return _stringify(record)
    return "; ".join(f"{k}={_stringify(v)}" for k, v in record.items())


def _build_result_row(raw_input: str, result) -> dict[str, str]:
    type_name = type(result).__name__
    try:
        record = result.to_record()
    except NotImplementedError:
        record = {}
    return {
        "input": raw_input,
        "type": type_name,
        "normalized": result.to_string(),
        "compact": result.compact_string(),
        "species": _stringify(getattr(result, "species_prefix", "")),
        "gene": _stringify(getattr(result, "gene_name", "")),
        "mhc_class": _stringify(getattr(result, "mhc_class", "")),
        "properties": _record_to_properties(record),
    }


def _build_error_row(raw_input: str, message: str) -> dict[str, str]:
    return {
        "input": raw_input,
        "type": "ParseError",
        "normalized": "",
        "compact": "",
        "species": "",
        "gene": "",
        "mhc_class": "",
        "properties": message,
    }


def _sanitize_cell(value: str) -> str:
    return value.replace("\t", " ").replace("\n", " ").strip()


def _render_tsv(rows: Sequence[dict[str, str]], include_header: bool) -> str:
    lines = []
    if include_header:
        lines.append("\t".join(COLUMN_NAMES))
    for row in rows:
        lines.append("\t".join(_sanitize_cell(row[column]) for column in COLUMN_NAMES))
    return "\n".join(lines)


def _render_table(rows: Sequence[dict[str, str]], include_header: bool) -> str:
    widths = {column: len(column) for column in COLUMN_NAMES}
    for row in rows:
        for column in COLUMN_NAMES:
            widths[column] = max(widths[column], len(_sanitize_cell(row[column])))

    lines = []
    if include_header:
        header = " | ".join(column.ljust(widths[column]) for column in COLUMN_NAMES)
        divider = "-+-".join("-" * widths[column] for column in COLUMN_NAMES)
        lines.extend([header, divider])

    for row in rows:
        row_text = " | ".join(
            _sanitize_cell(row[column]).ljust(widths[column]) for column in COLUMN_NAMES
        )
        lines.append(row_text)
    return "\n".join(lines)


def _build_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="mhcgnomes",
        description="Parse MHC strings and print a table with parsed properties.",
    )
    parser.add_argument(
        "names",
        nargs="*",
        help="MHC names to parse. If omitted, non-empty lines are read from stdin.",
    )
    parser.add_argument(
        "--default-species",
        default="HLA",
        help="Default species prefix used when the input is missing one (default: HLA).",
    )
    parser.add_argument(
        "--infer-class2-pairing",
        action="store_true",
        help="Infer canonical Class II alpha chain when only beta chain is given.",
    )
    parser.add_argument(
        "--max-allele-fields",
        type=int,
        default=None,
        help="If set, restrict parsed alleles to this many fields.",
    )
    parser.add_argument(
        "--format",
        choices=("table", "tsv", "json"),
        default="table",
        help="Output format (default: table).",
    )
    parser.add_argument(
        "--no-header",
        action="store_true",
        help="Omit header row in table/tsv output.",
    )
    parser.add_argument(
        "--strict",
        action="store_true",
        help="Exit with code 1 on the first parse error.",
    )
    return parser


def _collect_names(positional_names: Sequence[str]) -> list[str]:
    if positional_names:
        return [name for name in positional_names if name and name.strip()]
    if sys.stdin.isatty():
        return []
    return [line.strip() for line in sys.stdin if line.strip()]


def main(argv: Optional[Sequence[str]] = None) -> int:
    parser = _build_arg_parser()
    args = parser.parse_args(argv)

    names = _collect_names(args.names)
    if not names:
        parser.error("Provide one or more names as arguments or via stdin.")

    rows = []
    for name in names:
        try:
            result = parse(
                name,
                default_species=args.default_species,
                infer_class2_pairing=args.infer_class2_pairing,
                max_allele_fields=args.max_allele_fields,
                raise_on_error=args.strict,
            )
        except ParseError as err:
            print(f"ParseError: {err}", file=sys.stderr)
            return 1

        if result is None:
            message = f"Could not parse '{name}'"
            if args.strict:
                print(f"ParseError: {message}", file=sys.stderr)
                return 1
            rows.append(_build_error_row(name, message))
        else:
            rows.append(_build_result_row(name, result))

    if args.format == "json":
        print(json.dumps(rows, indent=2))
        return 0
    if args.format == "tsv":
        print(_render_tsv(rows, include_header=not args.no_header))
        return 0

    print(_render_table(rows, include_header=not args.no_header))
    return 0



from mhcgnomes import Species

from .common import eq_


def test_HLA():
    for hla in ["HLA", "hla"]:
        species = Species.get(hla)
        assert species is not None, f"Could not find Species for '{hla}'"
        eq_(species.common_species_name, "human")


def test_human():
    for human in ["human", "Human", "HUMAN"]:
        species = Species.get(human)
        assert species is not None, f"Could not find Species for '{human}'"
        eq_(species.common_species_name, "human")


def test_RT1():
    eq_(Species.get("RT1").prefix, "RT1")


def test_Rano():
    eq_(Species.get("Rano").prefix, "Rano")


def test_T2c_in_BoLA():
    cow = Species.get("BoLA")
    assert "T2c" in cow.known_alleles[None]


def test_known_allele_SLA_3_ydy01():
    pig = Species.get("SLA")
    eq_(pig.get_known_allele(gene_name="3", allele_name="ydy01"), ("3", "YDY01"))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import Optional, Union

from .gene import Gene
from .result_with_mhc_class import ResultWithMhcClass
from .species import Species


@dataclass(eq=False, repr=False, frozen=True, init=False)
class Class2Locus(ResultWithMhcClass):
    """
    Represents an MHC Class II locus (e.g., DR, DQ, DP).

    A Class II locus groups together the alpha and beta chain genes for a
    specific Class II molecule. For example, "HLA-DR" includes DRA (alpha)
    and DRB1, DRB3, DRB4, DRB5 (beta) genes.

    Parameters
    ----------
    species : Species
        The species this locus belongs to.
    name : str
        The locus name (e.g., "DR", "DQ", "DP").
    mhc_class : str, default "II"
        The MHC class (always "II" for Class2Locus).
    genes : Sequence[Gene], optional
        The genes belonging to this locus.
    raw_string : str, optional
        The original unparsed string.

    Attributes
    ----------
    species : Species
        The species this locus belongs to.
    name : str
        The locus name.
    genes : list of Gene
        The genes in this locus.
    alpha_chain_genes : list of Gene
        Alpha chain genes (names ending in "A").
    beta_chain_genes : list of Gene
        Beta chain genes (names ending in "B").

    Examples
    --------
    >>> from mhcgnomes import Class2Locus
    >>> locus = Class2Locus.get("HLA", "DR")
    >>> locus.name
    'DR'
    >>> locus.alpha_chain_gene_names
    ['DRA']
    """

    name: str = ""
    _genes: tuple[Gene, ...] = field(default_factory=tuple)

    def __init__(
        self,
        species: Species,
        name: str,
        mhc_class: str = "II",
        genes: Optional[Sequence[Gene]] = None,
        raw_string: Union[str, None] = None,
    ):
        ResultWithMhcClass.__init__(
            self, species=species, mhc_class=mhc_class, raw_string=raw_string
        )
        self._set_field(self, "name", name)
        self._set_field(self, "_genes", tuple(genes) if genes is not None else ())

    @property
    def locus_name(self):
        return self.name

    @property
    def genes(self):
        return list(self._genes)

    @classmethod
    def eq_field_names(cls):
        return ("species", "name")

    @classmethod
    def str_field_names(cls):
        return ("species", "name")

    def to_string(self, include_species=True, use_old_species_prefix=False):
        species_string = self.species.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )
        if species_string:
            return f"{species_string}-{self.name}"
        return self.name

    def compact_string(self, include_species=False, use_old_species_prefix=False):
        return self.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )

    @property
    def gene_names(self):
        return [g.name for g in self._genes]

    @classmethod
    def endswith_ignore_digits(cls, s1, s2):
        while s1 and s1[-1].isdigit():
            s1 = s1[:-1]
        return s1.endswith(s2)

    @property
    def alpha_chain_genes(self):
        results = []
        for g in self._genes:
            name = g.name
            if self.endswith_ignore_digits(name.lower(), "a"):
                results.append(g)
        return results

    @property
    def beta_chain_genes(self):
        results = []
        for g in self._genes:
            name = g.name
            if self.endswith_ignore_digits(name.lower(), "b"):
                results.append(g)
        return results

    @property
    def alpha_chain_gene_names(self):
        return [g.name for g in self.alpha_chain_genes]

    @property
    def beta_chain_gene_names(self):
        return [g.name for g in self.beta_chain_genes]

    @classmethod
    def get(cls, species_prefix, locus_name):
        if type(species_prefix) is Species:
            species = species_prefix
        else:
            species = Species.get(species_prefix)
            if species is None:
                return None

        raw_string = locus_name
        locus_name = species.find_matching_class2_locus_name(locus_name)
        if locus_name is None:
            return None

        gene_names = species.class2_locus_to_gene_names[locus_name]
        genes = []
        for gene_name in gene_names:
            gene = Gene.get(species, gene_name)
            if gene is None:
                continue
            genes.append(gene)
        return Class2Locus(species=species, name=locus_name, genes=genes, raw_string=raw_string)



"""
Tests for HLA serotype definitions based on WHO/IPD-IMGT/HLA Dictionary.

Source: https://www.ebi.ac.uk/ipd/imgt/hla/alleles/dictionary/

These tests document and verify the assumptions about serotype behavior based on
official WHO serological assignments from the IPD-IMGT/HLA Dictionary.

Key concepts:
-------------

1. SEROTYPES are serological antigen types - groups of HLA alleles that react
   with the same antibody in serological testing. Before molecular typing,
   HLA types were determined by antibody reactivity.

2. BROAD vs SPLIT SEROTYPES: Some original "broad" serotypes were later
   subdivided into "split" serotypes as more specific antisera were developed.
   For example:
   - A9 (broad) was split into A23 and A24
   - B15 (broad) was split into B62, B63, B70, B71, B72, B75, B76, B77

3. WHO ASSIGNMENT RULES: The WHO dictionary assigns serotypes based on actual
   serological reactivity:
   - Alleles that clearly react with split-specific antisera get the SPLIT
     serotype (e.g., A23, A24)
   - Alleles that react with the broad antiserum but don't clearly fit any
     split get the BROAD serotype (e.g., A9)
   - This means BROAD and SPLIT serotypes are DISJOINT sets - an allele
     is assigned to one or the other, not both.

4. PUBLIC EPITOPES (Bw4/Bw6): These are supertypic specificities defined by
   amino acid motifs at positions 77-83 of the HLA-B heavy chain. They are
   NOT traditional serotypes but cross-reactive epitopes shared by multiple
   serotypes.

5. WORKSHOP SEROTYPES: Names like "Aw68" or "DPw2" with "w" indicate workshop
   or provisional designations from international histocompatibility workshops.
   Many have since been officially recognized (Aw68 -> A68).
"""

from mhcgnomes import Species, parse
from mhcgnomes.serotype import Serotype


class TestSerotypeDataIntegrity:
    """Tests to verify the serotype data is loaded correctly."""

    def test_hla_serotypes_loaded(self):
        """HLA serotypes should be loaded from serotypes.yaml."""
        human = Species.get("HLA")
        assert human is not None
        assert hasattr(human, "serotypes")
        assert len(human.serotypes) > 100  # We have ~150 HLA serotypes

    def test_common_serotypes_exist(self):
        """Common clinical serotypes should all be present."""
        human = Species.get("HLA")
        common_serotypes = [
            # Class I - A locus
            "A1",
            "A2",
            "A3",
            "A11",
            "A23",
            "A24",
            "A25",
            "A26",
            "A29",
            "A30",
            "A31",
            "A32",
            "A33",
            "A34",
            "A36",
            "A43",
            "A66",
            "A68",
            "A69",
            "A74",
            "A80",
            # Class I - B locus
            "B7",
            "B8",
            "B13",
            "B18",
            "B27",
            "B35",
            "B37",
            "B38",
            "B39",
            "B41",
            "B42",
            "B44",
            "B45",
            "B46",
            "B47",
            "B48",
            "B49",
            "B50",
            "B51",
            "B52",
            "B53",
            "B54",
            "B55",
            "B56",
            "B57",
            "B58",
            "B59",
            "B60",
            "B61",
            "B62",
            "B63",
            "B64",
            "B65",
            "B67",
            "B73",
            "B78",
            "B81",
            # Class I - C locus
            "Cw1",
            "Cw2",
            "Cw3",
            "Cw4",
            "Cw5",
            "Cw6",
            "Cw7",
            "Cw8",
            # Class II - DR
            "DR1",
            "DR4",
            "DR7",
            "DR8",
            "DR9",
            "DR10",
            "DR11",
            "DR12",
            "DR13",
            "DR14",
            "DR15",
            "DR16",
            "DR17",
            "DR18",
            "DR51",
            "DR52",
            "DR53",
            # Class II - DQ
            "DQ2",
            "DQ4",
            "DQ5",
            "DQ6",
            "DQ7",
            "DQ8",
        ]
        for serotype in common_serotypes:
            assert serotype in human.serotypes, f"Missing common serotype: {serotype}"

    def test_serotypes_have_allele_lists(self):
        """Each serotype should have a list of alleles (may be empty for non-HLA)."""
        human = Species.get("HLA")
        for name, alleles in human.serotypes.items():
            assert isinstance(alleles, list), f"{name} alleles should be a list"


class TestBroadAndSplitSerotypes:
    """
    Tests for broad/split serotype relationships per WHO definitions.

    In HLA serology, some original "broad" serotypes were later subdivided into
    more specific "split" serotypes. The WHO dictionary assigns alleles to EITHER
    the broad OR a split serotype based on actual serological reactivity - they
    are DISJOINT sets, not hierarchical unions.

    This is the correct serological interpretation:
    - If an allele clearly reacts with split-specific antisera -> assigned to split
    - If an allele reacts with broad antiserum but not clearly with any split -> assigned to broad
    """

    # Known broad -> split relationships in HLA serology
    BROAD_SPLIT_RELATIONSHIPS = {
        # A locus
        "A9": ["A23", "A24"],
        "A10": ["A25", "A26", "A34", "A66"],
        "A19": ["A29", "A30", "A31", "A32", "A33", "A74"],
        "A28": ["A68", "A69"],
        # B locus
        "B5": ["B51", "B52"],
        "B12": ["B44", "B45"],
        "B14": ["B64", "B65"],
        "B15": ["B62", "B63", "B70", "B71", "B72", "B75", "B76", "B77"],
        "B16": ["B38", "B39"],
        "B17": ["B57", "B58"],
        "B21": ["B49", "B50"],
        "B22": ["B54", "B55", "B56"],
        "B40": ["B60", "B61"],
        # DR locus
        "DR2": ["DR15", "DR16"],
        "DR3": ["DR17", "DR18"],
        "DR5": ["DR11", "DR12"],
        "DR6": ["DR13", "DR14"],
    }

    def test_broad_and_split_serotypes_are_disjoint(self):
        """
        Broad and split serotypes should have disjoint allele sets per WHO rules.

        The WHO dictionary assigns each allele to exactly one serotype based on
        serological reactivity. An allele assigned to a split (e.g., A23) is NOT
        also assigned to its broad (A9).
        """
        human = Species.get("HLA")

        for broad, splits in self.BROAD_SPLIT_RELATIONSHIPS.items():
            broad_alleles = set(human.serotypes.get(broad, []))

            for split in splits:
                split_alleles = set(human.serotypes.get(split, []))
                overlap = broad_alleles & split_alleles

                assert not overlap, (
                    f"WHO rule violation: {broad} and {split} share alleles {overlap}. "
                    f"Broad and split serotypes should be disjoint per WHO assignments."
                )

    def test_broad_serotypes_contain_ambiguous_alleles(self):
        """
        Broad serotypes should contain alleles with ambiguous split assignment.

        Per WHO rules, alleles are assigned to the broad serotype when they
        react with the broad antiserum but don't clearly fit any split.
        """
        human = Species.get("HLA")

        # These broad serotypes should have at least some alleles per WHO dictionary
        # (alleles that don't clearly belong to any split)
        broad_serotypes_with_alleles = ["A9", "A10", "A19", "A28", "B14", "B15", "B40"]

        for broad in broad_serotypes_with_alleles:
            if broad in human.serotypes:
                alleles = human.serotypes[broad]
                assert len(alleles) > 0, (
                    f"Broad serotype {broad} should have alleles that don't clearly "
                    f"fit any split per WHO assignments"
                )

    def test_split_serotypes_exist_for_known_broads(self):
        """Split serotypes should exist when their broad serotype exists."""
        human = Species.get("HLA")

        for broad, splits in self.BROAD_SPLIT_RELATIONSHIPS.items():
            for split in splits:
                assert split in human.serotypes, (
                    f"Split serotype {split} (from {broad}) should exist"
                )


class TestPublicEpitopes:
    """
    Tests for Bw4 and Bw6 public epitopes (supertypic specificities).

    Bw4 and Bw6 are NOT in the WHO/IPD-IMGT/HLA Dictionary because they are
    public epitopes defined by amino acid motifs at positions 77-83 of the HLA
    heavy chain, not by traditional antibody reactivity.

    However, they ARE included in serotypes.yaml from supplemental source:
    https://hla.alleles.org/pages/antigens/bw4_bw6/

    Bw4-associated serotypes: B5, B13, B17, B27, B37, B38, B44, B47, B49, B51,
                              B52, B53, B57, B58, B59, B63, B77, A23, A24, A25, A32
    Bw6-associated serotypes: B7, B8, B14, B18, B22, B35, B39, B40, B41, B42,
                              B45, B46, B48, B50, B54, B55, B56, B60, B61, B62,
                              B64, B65, B67, B70, B71, B72, B73, B75, B76, B78,
                              B81, B82
    """

    def test_bw4_exists_with_alleles(self):
        """Bw4 public epitope should be defined with representative alleles."""
        human = Species.get("HLA")
        assert "Bw4" in human.serotypes
        bw4_alleles = human.serotypes["Bw4"]
        assert len(bw4_alleles) > 0

    def test_bw6_exists_with_alleles(self):
        """Bw6 public epitope should be defined with representative alleles."""
        human = Species.get("HLA")
        assert "Bw6" in human.serotypes
        bw6_alleles = human.serotypes["Bw6"]
        assert len(bw6_alleles) > 0

    def test_bw4_contains_expected_alleles(self):
        """Bw4 should contain alleles from Bw4-positive serotypes."""
        human = Species.get("HLA")
        bw4_alleles = set(human.serotypes["Bw4"])

        # Representative alleles that carry Bw4 epitope
        expected = ["B*2705", "B*5101", "B*5701", "B*4402"]
        for allele in expected:
            assert allele in bw4_alleles, f"Expected {allele} in Bw4"

    def test_bw6_contains_expected_alleles(self):
        """Bw6 should contain alleles from Bw6-positive serotypes."""
        human = Species.get("HLA")
        bw6_alleles = set(human.serotypes["Bw6"])

        # Representative alleles that carry Bw6 epitope
        expected = ["B*0702", "B*0801", "B*3501"]
        for allele in expected:
            assert allele in bw6_alleles, f"Expected {allele} in Bw6"

    def test_bw4_and_bw6_are_disjoint(self):
        """Bw4 and Bw6 epitopes should be mutually exclusive."""
        human = Species.get("HLA")
        bw4 = set(human.serotypes["Bw4"])
        bw6 = set(human.serotypes["Bw6"])
        overlap = bw4 & bw6
        assert not overlap, f"Bw4 and Bw6 should be disjoint, overlap: {overlap}"

    def test_bw4_positive_serotypes_also_exist(self):
        """Individual Bw4-positive serotypes should also exist."""
        human = Species.get("HLA")
        assert "B27" in human.serotypes
        assert "B44" in human.serotypes

    def test_bw6_positive_serotypes_also_exist(self):
        """Individual Bw6-positive serotypes should also exist."""
        human = Species.get("HLA")
        assert "B7" in human.serotypes
        assert "B8" in human.serotypes


class TestWorkshopSerotypes:
    """
    Tests for workshop (provisional) serotype designations.

    The "w" in serotype names like "Aw68", "Bw4", "Cw1", "DPw2" originally
    indicated a "workshop" or provisional designation from international
    histocompatibility workshops.

    Current status:
    - Aw68, Aw69: Preserved as aliases for A68, A69 (for backward compatibility)
    - Bw4, Bw6: Public epitopes (see TestPublicEpitopes)
    - Cw serotypes: C locus serotypes traditionally used "w" (Cw1, Cw2, etc.)
    - DPw serotypes: DP workshop serotypes (DPw1-DPw6)
    """

    def test_modern_a68_a69_exist(self):
        """A68 and A69 serotypes should exist (modern/official names)."""
        human = Species.get("HLA")
        assert "A68" in human.serotypes
        assert "A69" in human.serotypes

    def test_old_workshop_aw68_aw69_exist_as_aliases(self):
        """
        Old Aw68/Aw69 names should exist for backward compatibility.

        These are preserved as aliases pointing to the same alleles as A68/A69.
        Source: https://hla.alleles.org/pages/antigens/previous_equivalents/
        """
        human = Species.get("HLA")
        assert "Aw68" in human.serotypes
        assert "Aw69" in human.serotypes
        # Should have alleles
        assert len(human.serotypes["Aw68"]) > 0
        assert len(human.serotypes["Aw69"]) > 0

    def test_aw68_and_a68_have_same_alleles(self):
        """Aw68 and A68 should contain the same alleles (aliases)."""
        human = Species.get("HLA")
        aw68 = set(human.serotypes["Aw68"])
        a68 = set(human.serotypes["A68"])
        # They should be equivalent
        assert aw68 == a68, f"Aw68 and A68 should match: {aw68} vs {a68}"

    def test_cw_serotypes_use_w_prefix(self):
        """C locus serotypes should use traditional Cw naming."""
        human = Species.get("HLA")
        cw_serotypes = ["Cw1", "Cw2", "Cw3", "Cw4", "Cw5", "Cw6", "Cw7", "Cw8"]
        for cw in cw_serotypes:
            assert cw in human.serotypes, f"Expected {cw} serotype"

    def test_dpw_serotypes_exist(self):
        """DP workshop serotypes should exist (preserved for compatibility)."""
        human = Species.get("HLA")
        dpw_serotypes = ["DPw1", "DPw2", "DPw3", "DPw4", "DPw5", "DPw6"]
        for dpw in dpw_serotypes:
            assert dpw in human.serotypes, f"Expected {dpw} serotype"


class TestSerotypeToAlleleMapping:
    """Tests for the serotype-to-allele mapping accuracy."""

    def test_a2_contains_common_alleles(self):
        """A2 serotype should contain common A*02 alleles."""
        human = Species.get("HLA")
        a2_alleles = set(human.serotypes["A2"])

        # A*02:01 is the most common A2 allele worldwide
        assert "A*0201" in a2_alleles, "A*02:01 is the archetypal A2 allele"

        # Other common A2 alleles
        common_a2 = ["A*0201", "A*0202", "A*0205", "A*0206"]
        for allele in common_a2:
            assert allele in a2_alleles, f"Expected {allele} in A2 serotype"

    def test_b27_contains_common_alleles(self):
        """B27 serotype should contain B*27 alleles (important for disease association)."""
        human = Species.get("HLA")
        b27_alleles = set(human.serotypes["B27"])

        # B*27:05 is the most common B27 subtype associated with ankylosing spondylitis
        assert "B*2705" in b27_alleles, "B*27:05 is the classic B27 disease-associated allele"

    def test_dr4_contains_common_alleles(self):
        """DR4 serotype should contain DRB1*04 alleles."""
        human = Species.get("HLA")
        dr4_alleles = set(human.serotypes["DR4"])

        common_dr4 = ["DRB1*0401", "DRB1*0404", "DRB1*0405"]
        for allele in common_dr4:
            assert allele in dr4_alleles, f"Expected {allele} in DR4 serotype"

    def test_allele_format_consistency(self):
        """Alleles should use consistent two-field format (e.g., A*0201 not A*02:01)."""
        human = Species.get("HLA")

        for serotype, alleles in human.serotypes.items():
            for allele in alleles:
                # Should not contain colons (we use compact format)
                assert ":" not in allele or allele.count(":") <= 1, (
                    f"Allele {allele} in {serotype} should use compact format"
                )


class TestSerotypeSubtypes:
    """Tests for serotype subtypes (e.g., A203, A2403, B703)."""

    def test_a203_is_distinct_from_a2(self):
        """A203 is a distinct serotype, not the same as A2."""
        human = Species.get("HLA")
        assert "A203" in human.serotypes
        assert "A2" in human.serotypes

        # A203 should contain A*02:03 specifically
        a203_alleles = human.serotypes["A203"]
        assert "A*0203" in a203_alleles

        # A*02:03 should NOT be in A2
        a2_alleles = human.serotypes["A2"]
        assert "A*0203" not in a2_alleles, "A*02:03 has A203 serotype, not A2"

    def test_b703_is_distinct_from_b7(self):
        """B703 is a distinct serotype from B7."""
        human = Species.get("HLA")
        assert "B703" in human.serotypes
        assert "B7" in human.serotypes

        b703_alleles = set(human.serotypes["B703"])
        b7_alleles = set(human.serotypes["B7"])

        # These should be disjoint
        overlap = b703_alleles & b7_alleles
        assert not overlap, f"B703 and B7 should be disjoint, overlap: {overlap}"


class TestSerotypeParsing:
    """Tests for parsing serotype strings."""

    def test_parse_hla_a2(self):
        """Should parse HLA-A2 as a serotype."""
        result = parse("HLA-A2")
        assert result is not None
        assert isinstance(result, Serotype)
        assert result.name == "A2"
        assert result.species.prefix == "HLA"

    def test_parse_hla_b27(self):
        """Should parse HLA-B27 as a serotype."""
        result = parse("HLA-B27")
        assert result is not None
        assert isinstance(result, Serotype)
        assert result.name == "B27"

    def test_parse_hla_dr4(self):
        """Should parse HLA-DR4 as a serotype."""
        result = parse("HLA-DR4")
        assert result is not None
        assert isinstance(result, Serotype)
        assert result.name == "DR4"

    def test_parse_hla_cw7(self):
        """Should parse HLA-Cw7 as a serotype."""
        result = parse("HLA-Cw7")
        assert result is not None
        assert isinstance(result, Serotype)
        assert result.name == "Cw7"

    def test_parse_hla_dpw4(self):
        """Should parse HLA-DPw4 as a serotype."""
        result = parse("HLA-DPw4")
        assert result is not None
        assert isinstance(result, Serotype)
        assert result.name == "DPw4"

    def test_parse_hla_dr51(self):
        """Should parse HLA-DR51 as a serotype."""
        result = parse("HLA-DR51")
        assert result is not None
        assert isinstance(result, Serotype)
        assert result.name == "DR51"

    def test_parse_hla_bw4(self):
        """Should parse HLA-Bw4 as a serotype (public epitope)."""
        result = parse("HLA-Bw4")
        assert result is not None
        assert isinstance(result, Serotype)
        assert result.name == "Bw4"
        assert len(result.alleles) > 0

    def test_parse_hla_aw68(self):
        """Should parse HLA-Aw68 as a serotype (workshop name)."""
        result = parse("HLA-Aw68")
        assert result is not None
        assert isinstance(result, Serotype)
        assert result.name == "Aw68"
        assert len(result.alleles) > 0


class TestEdgeCases:
    """
    Tests for edge cases in serotype definitions.

    These cover unusual but valid serotype patterns from the WHO dictionary:
    - Variant serotypes with lowercase suffixes (B27c, DR3d, etc.)
    - B17 subsplits (B17.1, B17.2, B17.3)
    - Alleles with dual serological reactivity
    - Case-insensitive parsing
    """

    def test_variant_serotypes_with_lowercase_suffix(self):
        """
        Variant serotypes with lowercase suffixes should be parseable.

        These represent variant serological reactivity patterns assigned by WHO.
        """
        human = Species.get("HLA")
        variants = ["B27c", "B48c", "Cw4c", "DQ6c", "DR3d"]
        for variant in variants:
            assert variant in human.serotypes, f"Expected variant serotype {variant}"
            assert len(human.serotypes[variant]) > 0, f"{variant} should have alleles"

    def test_parse_variant_serotypes(self):
        """Should parse variant serotypes with lowercase suffixes."""
        result = parse("HLA-B27c")
        assert result is not None
        assert result.name == "B27c"

        result = parse("HLA-DR3d")
        assert result is not None
        assert result.name == "DR3d"

    def test_b17_subsplits(self):
        """
        B17 has historical subsplits (B17.1, B17.2, B17.3) in addition to B57/B58.

        These represent finer serological distinctions within B17.
        """
        human = Species.get("HLA")
        subsplits = ["B17.1", "B17.2", "B17.3"]
        for subsplit in subsplits:
            assert subsplit in human.serotypes, f"Expected B17 subsplit {subsplit}"
            assert len(human.serotypes[subsplit]) > 0

    def test_parse_b17_subsplits(self):
        """Should parse B17 subsplit serotypes."""
        result = parse("HLA-B17.1")
        assert result is not None
        assert result.name == "B17.1"

    def test_dual_reactivity_alleles_documented(self):
        """
        Some alleles have dual serological reactivity per WHO.

        These alleles legitimately appear in multiple serotypes because they
        react with antisera for both serotypes.
        """
        human = Species.get("HLA")

        # A*24:18 reacts with both A24 and A3 antisera
        a24_alleles = set(human.serotypes.get("A24", []))
        a3_alleles = set(human.serotypes.get("A3", []))
        assert "A*2418" in a24_alleles or "A*2418" in a3_alleles

        # DRB1*11:16 reacts with both DR11 and DR13 antisera
        dr11_alleles = set(human.serotypes.get("DR11", []))
        dr13_alleles = set(human.serotypes.get("DR13", []))
        assert "DRB1*1116" in dr11_alleles or "DRB1*1116" in dr13_alleles

    def test_case_insensitive_parsing(self):
        """Serotype parsing should be case-insensitive."""
        # All should parse to the same serotype
        for name in ["HLA-A2", "hla-a2", "HLA-a2", "A2", "a2"]:
            result = parse(name)
            assert result is not None, f"Failed to parse {name}"
            assert result.name == "A2", f"{name} should parse to A2"

    def test_ambiguous_serotype_names_exist(self):
        """
        WHO serotype names that look like alleles should still exist in serotypes.

        The WHO dictionary defines serotypes like B3901, B5102, etc. Although these
        names look like allele patterns (B*39:01, B*51:02), they are legitimate WHO
        serotype designations and should be included in the serotype data.

        The parser handles ambiguity by prioritizing allele interpretation over
        serotype interpretation when both are valid candidates.
        """
        human = Species.get("HLA")
        # These WHO serotype names should exist (they are legitimate serotypes)
        ambiguous_names = ["B3901", "B3902", "B5102", "B5103", "B4005", "B2708"]
        for name in ambiguous_names:
            assert name in human.serotypes, f"{name} should exist as a valid WHO serotype"

    def test_allele_like_names_parse_as_alleles(self):
        """
        Names that look like alleles should parse as alleles, not serotypes.

        Even though B3901, B5102, etc. are valid WHO serotypes, when a user writes
        "B3901" they almost always mean the allele B*39:01, not the serotype.
        The parser prioritizes allele interpretation over serotype interpretation.
        """
        from mhcgnomes.allele import Allele

        # All these should parse as alleles (not serotypes, even though B3901 is a valid serotype)
        for name in ["B3901", "B*3901", "B*39:01", "HLA-B3901"]:
            result = parse(name)
            assert isinstance(result, Allele), f"{name} should parse as allele, not serotype"
            assert "39:01" in result.to_string() or "3901" in result.to_string()


class TestNonHumanSerotypes:
    """Tests for non-human species serotypes."""

    def test_patr_serotypes_exist(self):
        """Chimpanzee (Patr) serotypes should be defined."""
        patr = Species.get("Patr")
        assert patr is not None
        assert hasattr(patr, "serotypes")
        # Patr has DR serotypes defined (even if empty allele lists)
        assert "DR1" in patr.serotypes

    def test_mamu_serotypes_exist(self):
        """Rhesus macaque (Mamu) serotypes should be defined."""
        mamu = Species.get("Mamu")
        assert mamu is not None
        assert hasattr(mamu, "serotypes")

    def test_bola_serotypes_exist(self):
        """Bovine (BoLA) serotypes should be defined."""
        bola = Species.get("BoLA")
        assert bola is not None
        assert hasattr(bola, "serotypes")



"""
Tests for batch-added species from the mhcseqs P2/P5 requests.
Covers raptors, cranes, passerines, additional fish, amphibians,
sharks, tuatara, and taxonomy reclassifications.
"""

import pytest

from mhcgnomes import Species, parse

from .common import eq_

BATCH_SPECIES = [
    # Raptors
    ("Fape", "Falco peregrinus"),
    ("Fati", "Falco tinnunculus"),
    ("Aqch", "Aquila chrysaetos"),
    ("ButeoBute", "Buteo buteo"),
    ("Ciae", "Circus aeruginosus"),
    ("Gyfu", "Gyps fulvus"),
    # Cranes
    ("Grja", "Grus japonensis"),
    ("Anan", "Antigone antigone"),
    ("Bare", "Balearica regulorum"),
    # Passerines
    ("Pado", "Passer domesticus"),
    ("Coco", "Corvus cornix"),
    ("Tume", "Turdus merula"),
    ("Laco", "Lanius collurio"),
    ("Emci", "Emberiza citrinella"),
    # Shorebirds
    ("Anni", "Anarhynchus nivosus"),
    # Fish
    ("Gaac", "Gasterosteus aculeatus"),
    ("Icpu", "Ictalurus punctatus"),
    ("Hiab", "Hippocampus abdominalis"),
    ("Caau", "Carassius auratus"),
    ("Clma", "Clarias magur"),
    # Amphibians
    ("Lica", "Lithobates catesbeianus"),
    ("Licl", "Lithobates clamitans"),
    ("Amme", "Ambystoma mexicanum"),
    ("Amti", "Ambystoma tigrinum"),
    # Tuatara
    ("Sppu", "Sphenodon punctatus"),
    # Marsupial
    ("Noeu", "Notamacropus eugenii"),
    # Sharks
    ("Gici", "Ginglymostoma cirratum"),
    ("Hefr", "Heterodontus francisci"),
    ("Sqac", "Squalus acanthias"),
    ("Trsc", "Triakis scyllium"),
]


@pytest.mark.parametrize("prefix,latin", BATCH_SPECIES)
def test_batch_species_parseable(prefix, latin):
    species = Species.get(prefix)
    assert species is not None, f"Species.get({prefix!r}) returned None"
    eq_(species.name, latin)
    eq_(parse(prefix, raise_on_error=True), species)
    # Latin name should also work
    eq_(Species.get(latin), species)


# Taxonomy reclassification aliases
RECLASSIFICATION_ALIASES = [
    ("Gran", "Antigone antigone"),  # Grus → Antigone
    ("Chni", "Anarhynchus nivosus"),  # Charadrius → Anarhynchus
    ("Raca", "Lithobates catesbeianus"),  # Rana → Lithobates
    ("Racl", "Lithobates clamitans"),  # Rana → Lithobates
    ("Maeu", "Notamacropus eugenii"),  # Macropus → Notamacropus
]


@pytest.mark.parametrize("old_prefix,latin", RECLASSIFICATION_ALIASES)
def test_old_genus_prefix_resolves_to_current_species(old_prefix, latin):
    species = Species.get(old_prefix)
    assert species is not None, f"Old prefix {old_prefix!r} did not resolve"
    eq_(species.name, latin)



from mhcgnomes import Allele
from mhcgnomes.standard_format import parse_standard_allele_format

from .common import eq_


def test_parse_standard_allele_format_HLA_A_02_01():
    result = parse_standard_allele_format(seq="HLA-A*02:01")
    eq_(result, Allele.get("HLA", "A", "02", "01"))


def test_parse_standard_allele_format_HLA_A_02_01_01():
    result = parse_standard_allele_format(seq="HLA-A*02:01:01")
    eq_(result, Allele.get("HLA", "A", "02", "01", "01"))


def test_parse_standard_allele_format_HLA_A_02_01_01_01():
    result = parse_standard_allele_format(seq="HLA-A*02:01:01:01")
    eq_(result, Allele.get("HLA", "A", "02", "01", "01", "01"))


def test_parse_standard_allele_format_HLA_A_02_01_01_02L():
    result = parse_standard_allele_format(seq="HLA-A*02:01:01:02L")
    eq_(result, Allele.get("HLA", "A", "02", "01", "01", "02", annotation="L"))


def test_parse_standard_allele_format_DLA_88_021_01():
    result = parse_standard_allele_format(seq="DLA-88*01:01")
    eq_(result, Allele.get("DLA", "88", "01", "01"))


def test_parse_standard_allele_format_A_02_01():
    result = parse_standard_allele_format(seq="A*02:01", default_species="HLA")
    eq_(result, Allele.get("HLA", "A", "02", "01"))



"""
Tests for Species identity model where latin name is the canonical identity
and prefixes/common names are aliases that may be ambiguous.
"""

from mhcgnomes import Allele, Gene, Species, parse

from .common import eq_

# ---------------------------------------------------------------------------
# 1. Species identity via latin name
# ---------------------------------------------------------------------------


def test_species_get_by_latin_name_human():
    species = Species.get_by_latin_name("Homo sapiens")
    assert species is not None
    eq_(species.prefix, "HLA")


def test_species_get_by_latin_name_mouse():
    species = Species.get_by_latin_name("Mus musculus")
    assert species is not None
    eq_(species.prefix, "H2")


def test_species_get_by_latin_name_chicken():
    species = Species.get_by_latin_name("Gallus gallus")
    assert species is not None
    eq_(species.prefix, "Gaga")


def test_species_get_by_latin_name_unknown_returns_none():
    assert Species.get_by_latin_name("Nonexistus fictionalus") is None


def test_species_latin_name_property():
    species = Species.get_by_latin_name("Homo sapiens")
    eq_(species.latin_name, "Homo sapiens")
    eq_(species.name, "Homo sapiens")


def test_species_to_record_includes_latin_name():
    species = Species.get_by_latin_name("Homo sapiens")
    record = species.to_record()
    assert "species_latin_name" in record
    eq_(record["species_latin_name"], "Homo sapiens")
    # Backwards compatibility
    assert "species_name" in record
    eq_(record["species_name"], "Homo sapiens")


# ---------------------------------------------------------------------------
# 2. Unique alias regression tests
# ---------------------------------------------------------------------------


def test_species_get_unique_prefix_still_works():
    for prefix, expected_name in [
        ("HLA", "Homo sapiens"),
        ("H2", "Mus musculus"),
        ("BoLA", "Bos sp."),
    ]:
        species = Species.get(prefix)
        assert species is not None, f"Species.get({prefix!r}) returned None"
        eq_(species.name, expected_name)


def test_species_get_unique_common_name_still_works():
    for name in ["human", "mouse"]:
        species = Species.get(name)
        assert species is not None, f"Species.get({name!r}) returned None"


def test_parse_unique_prefix_strings_unchanged():
    """Representative existing cases still parse identically."""
    examples = [
        ("HLA-A*02:01", Allele),
        ("Gaga-BF1", Gene),
        ("Dare-UBA", Gene),
    ]
    for raw, expected_type in examples:
        result = parse(raw, raise_on_error=True)
        assert isinstance(result, expected_type), (
            f"{raw} parsed as {type(result)}, expected {expected_type}"
        )


# ---------------------------------------------------------------------------
# 3. Ambiguity tests using real runtime collision (Bubu)
# ---------------------------------------------------------------------------


def test_species_get_bubu_resolves_to_buffalo():
    """Bubu is now uniquely owned by water buffalo (eagle-owl is BuboBubo)."""
    species = Species.get("Bubu")
    assert species is not None
    eq_(species.name, "Bubalus bubalis")


def test_parse_bubu_dqa_resolves_water_buffalo():
    """Bubu-DQA should resolve to Bubalus bubalis."""
    result = parse("Bubu-DQA", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.name, "Bubalus bubalis")


def test_parse_bubobubo_dab1_resolves_eagle_owl():
    """BuboBubo-DAB1 should resolve to eagle-owl via long prefix."""
    result = parse("BuboBubo-DAB1", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.name, "Bubo bubo")


def test_parse_bubu_allele_resolves_by_gene_context():
    """Bubu-DQA*01:01 should resolve to buffalo."""
    result = parse("Bubu-DQA*01:01", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.name, "Bubalus bubalis")


def test_species_get_by_latin_name_buffalo():
    """Direct latin name lookup bypasses ambiguity."""
    species = Species.get_by_latin_name("Bubalus bubalis")
    assert species is not None
    eq_(species.prefix, "Bubu")


def test_species_get_by_latin_name_eagle_owl():
    """Direct latin name lookup bypasses ambiguity."""
    species = Species.get_by_latin_name("Bubo bubo")
    assert species is not None
    eq_(species.prefix, "BuboBubo")


def test_long_prefix_always_parseable():
    """The 5+5 long prefix should always work for parsing any species."""
    for latin, long_prefix in [
        ("Bubo bubo", "BuboBubo"),
        ("Chrysemys picta", "ChrysPicta"),
        ("Gavialis gangeticus", "GaviaGange"),
        ("Casuarius casuarius", "CasuaCasua"),
        ("Caretta caretta", "CaretCaret"),
    ]:
        species = Species.get(long_prefix)
        assert species is not None, f"Species.get({long_prefix!r}) returned None"
        eq_(species.name, latin)


def test_long_prefix_works_for_allele_parsing():
    """HomoSapie-A*02:01 should parse as HLA-A*02:01."""
    result = parse("HomoSapie-A*02:01", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")
    eq_(result.gene.name, "A")
    eq_(result.allele_fields, ("02", "01"))


def test_long_prefix_works_for_compact_allele():
    """HomoSapie-A0201 should also parse."""
    result = parse("HomoSapie-A0201", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")
    eq_(result.allele_fields, ("02", "01"))


def test_full_latin_name_concatenated_works():
    """HomoSapiens-A*02:01 should parse (full latin name, no truncation)."""
    result = parse("HomoSapiens-A*02:01", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")
    eq_(result.gene.name, "A")
    eq_(result.allele_fields, ("02", "01"))


def test_full_latin_name_concatenated_various_species():
    """Full concatenated latin names should work for any species."""
    for concat, expected_prefix in [
        ("DanioRerio-UBA", "Dare"),
        ("GallusGallus-BF1", "Gaga"),
        ("MusMusculus-K", "H2"),
    ]:
        result = parse(concat, raise_on_error=True)
        assert result is not None, f"parse({concat!r}) returned None"
        eq_(result.species.prefix, expected_prefix)


def test_latin_name_with_space_works():
    """'Homo sapiens-A*02:01' with a space in the species name should work."""
    result = parse("Homo sapiens-A*02:01", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")


# ---------------------------------------------------------------------------
# 6. Failure modes for latin name / long prefix parsing
# ---------------------------------------------------------------------------


def test_nonexistent_latin_name_returns_none():
    assert parse("FictusSpecius-A*02:01", raise_on_error=False) is None


def test_partial_latin_name_too_short_returns_none():
    """Just 'Homo' should not resolve to a full allele."""
    assert parse("Homo-A*02:01", raise_on_error=False) is None


def test_misspelled_latin_name_returns_none():
    assert parse("HomoSapeins-A*02:01", raise_on_error=False) is None


def test_wrong_gene_for_species_returns_none():
    """HomoSapiens-BF1 — BF1 is a chicken gene, not human."""
    assert parse("HomoSapiens-BF1", raise_on_error=False) is None


# ---------------------------------------------------------------------------
# 4. Default species accepts latin name
# ---------------------------------------------------------------------------


def test_default_species_accepts_latin_name():
    """Parsing with default_species as latin name should work."""
    result = parse("A*02:01", default_species="Homo sapiens")
    assert result is not None
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")



from mhcgnomes import Species, parse

from .common import eq_


def test_species_lineage_helpers():
    bos = Species.get("Bos sp.")
    bota = Species.get("Bos taurus")
    galliformes = Species.get("Galliformes sp.")
    gaga = Species.get("Gallus gallus")

    assert bos is not None
    assert bota is not None
    assert galliformes is not None
    assert gaga is not None

    assert bos.is_parent_of(bota)
    assert bota.is_child_of(bos)
    assert bos.is_ancestor_of(bota)
    assert bota.is_descendant_of(bos)
    assert galliformes.is_ancestor_of(gaga)
    assert gaga.is_descendant_of(galliformes)


def test_species_strict_exact_match():
    result = parse("HLA-A*02:01", species="Homo sapiens")
    eq_(result.species.name, "Homo sapiens")


def test_species_strict_bola_converts_to_bos_taurus():
    result = parse("BoLA-DRB3*01:01", species="Bos taurus")
    eq_(result.species.name, "Bos taurus")
    eq_(result.to_string(), "Bota-DRB3*01:01")


def test_species_strict_sla_converts_to_sus_scrofa():
    result = parse("SLA-1*01:01", species="Sus scrofa")
    eq_(result.species.name, "Sus scrofa")
    eq_(result.to_string(), "Susc-1*01:01")


def test_species_strict_dla_converts_to_canis_lupus():
    result = parse("DLA-DRB1", species="Canis lupus")
    eq_(result.species.name, "Canis lupus")
    eq_(result.to_string(), "Calu-DRB1")


def test_species_strict_bola_with_bos_sp():
    result = parse("BoLA-DRB3*01:01", species="Bos sp.")
    eq_(result.species.name, "Bos sp.")
    eq_(result.to_string(), "BoLA-DRB3*01:01")


def test_species_strict_hla_with_bos_taurus_errors():
    assert parse("HLA-A*02:01", species="Bos taurus", raise_on_error=False) is None


def test_species_strict_bola_with_human_errors():
    assert parse("BoLA-DRB3*01:01", species="Homo sapiens", raise_on_error=False) is None


def test_species_strict_bola_with_sheep_errors():
    assert parse("BoLA-DRB3*01:01", species="Ovis sp.", raise_on_error=False) is None


def test_species_strict_taxonomic_node_does_not_convert_to_descendant():
    assert parse("Galliformes", species="Gallus gallus", raise_on_error=False) is None


def test_species_strict_child_result_not_cast_up_to_parent():
    assert parse("Gaga-BF1", species="Galliformes sp.", raise_on_error=False) is None



from mhcgnomes import Allele, Gene, Pair, Species, parse

from .common import eq_


def test_parse_trout_species_Onmy():
    expected = Species.get("Onmy")
    assert expected is not None
    result = parse("Onmy", raise_on_error=True)
    assert result is not None
    eq_(result, expected)


def test_parse_trout_gene_Onmy_DAB():
    result = parse("Onmy-DAB", raise_on_error=True)
    expected = Gene.get("Onmy", "DAB")
    assert expected is not None
    eq_(result, expected)


def test_parse_trout_allele_Onmy_DAB_0501():
    result = parse("Onmy-DAB*0501", raise_on_error=True)
    expected = Allele.get("Onmy", "DAB", ["05", "01"])
    assert expected is not None
    eq_(result, expected)


def test_parse_trout_allele_Onmy_DAB_050101():
    result = parse("Onmy-DAB*050101", raise_on_error=True)
    expected = Allele.get("Onmy", "DAB", ["05", "01", "01"])
    assert expected is not None
    eq_(result, expected)


def test_parse_trout_gene_Onmy_DAA1():
    """DAA1 is now a real gene on Actinopterygii."""
    result = parse("Onmy-DAA1", raise_on_error=True)
    expected = Gene.get("Onmy", "DAA1")
    assert expected is not None
    eq_(result, expected)


def test_parse_trout_allele_Onmy_DAA1_01_01():
    result = parse("Onmy-DAA1*01:01", raise_on_error=True)
    expected = Allele.get("Onmy", "DAA1", ["01", "01"])
    assert expected is not None
    eq_(result, expected)


def test_parse_salmon_gene_Sasa_DAA1():
    result = parse("Sasa-DAA1", raise_on_error=True)
    expected = Gene.get("Sasa", "DAA1")
    assert expected is not None
    eq_(result, expected)


def test_parse_salmon_allele_Sasa_DAA1_01_01():
    result = parse("Sasa-DAA1*01:01", raise_on_error=True)
    expected = Allele.get("Sasa", "DAA1", ["01", "01"])
    assert expected is not None
    eq_(result, expected)


def test_parse_grass_carp_B2M_1_ii_star_sep():
    expected = Allele.get("Ctid", "B2M-1", "ii")
    result = parse("Ctid-B2M-1*ii")
    eq_(result, expected)


def test_parse_grass_carp_B2M_1_ii_dash_sep():
    expected = Allele.get("Ctid", "B2M-1", "ii")
    result = parse("Ctid-B2M-1-ii")
    eq_(result, expected)


def test_parse_grass_carp_mhc1_pair_UAA_B2M_1_ii_dash_sep():
    expected = Pair(alpha=Gene.get("Ctid", "UAA"), beta=Allele.get("Ctid", "B2M-1", "ii"))
    result = parse("Ctid-UAA/B2M-1-ii")
    eq_(result, expected)


def test_parse_grass_carp_mhc1_pair_UAA_B2M_1_ii_star_sep():
    expected = Pair(alpha=Gene.get("Ctid", "UAA"), beta=Allele.get("Ctid", "B2M-1", "ii"))
    result = parse("Ctid-UAA/B2M-1*ii")
    eq_(result, expected)


def test_parse_olive_flounder_species_Paol():
    expected = Species.get("Paol")
    assert expected is not None
    eq_(parse("Paol", raise_on_error=True), expected)


def test_parse_zebrafish_species_Dare():
    expected = Species.get("Dare")
    assert expected is not None
    eq_(parse("Dare", raise_on_error=True), expected)


def test_parse_zebrafish_UBA_gene_and_allele():
    expected_gene = Gene.get("Dare", "UBA")
    expected_allele = Allele.get("Dare", "UBA", "01")
    assert expected_gene is not None
    assert expected_allele is not None
    eq_(expected_gene.mhc_class, "Ia")
    eq_(parse("Dare-UBA", raise_on_error=True), expected_gene)
    eq_(parse("Dare-UBA*01", raise_on_error=True), expected_allele)


def test_parse_common_carp_species_Cyca():
    expected = Species.get("Cyca")
    assert expected is not None
    eq_(parse("Cyca", raise_on_error=True), expected)


def test_parse_common_carp_family_level_genes():
    expected_classes = {"UA": "I", "ZE": "I", "DXA": "IIa", "DAB": "IIa"}
    for gene_name in ["UA", "ZE", "DXA", "DAB"]:
        expected = Gene.get("Cyca", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, expected_classes[gene_name])
        eq_(parse(f"Cyca-{gene_name}", raise_on_error=True), expected)


def test_parse_common_carp_numbered_loci_and_aliases():
    """DAB1-3 are now real genes on Actinopterygii. DXA1/2 and UA1 remain aliases."""
    examples = [
        ("Cyca-UA1", Gene.get("Cyca", "UA")),
        ("Cyca-DXA1", Gene.get("Cyca", "DXA")),
        ("Cyca-DXA2", Gene.get("Cyca", "DXA")),
        ("Cyca-DAB1", Gene.get("Cyca", "DAB1")),
        ("Cyca-DAB2", Gene.get("Cyca", "DAB2")),
        ("Cyca-DAB3", Gene.get("Cyca", "DAB3")),
        ("Cyca-DAB4", Gene.get("Cyca", "DAB")),
    ]
    for raw_string, expected in examples:
        assert expected is not None
        eq_(parse(raw_string, raise_on_error=True), expected)


def test_parse_common_carp_alias_alleles():
    examples = [
        ("Cyca-UA1*01", Allele.get("Cyca", "UA", "01")),
        ("Cyca-DXA1*01", Allele.get("Cyca", "DXA", "01")),
        ("Cyca-DXA2*01", Allele.get("Cyca", "DXA", "01")),
        ("Cyca-DAB1*01", Allele.get("Cyca", "DAB1", "01")),
        ("Cyca-DAB3*01", Allele.get("Cyca", "DAB3", "01")),
        ("Cyca-DAB4*01", Allele.get("Cyca", "DAB", "01")),
        ("Cyca-ZE*01:01", Allele.get("Cyca", "ZE", "01", "01")),
    ]
    for raw_string, expected in examples:
        assert expected is not None
        eq_(parse(raw_string, raise_on_error=True), expected)


def test_parse_nile_tilapia_species_Orni():
    expected = Species.get("Orni")
    assert expected is not None
    eq_(parse("Orni", raise_on_error=True), expected)


def test_parse_nile_tilapia_genes_and_example_alleles():
    examples = [
        ("DAA", ("05", "01")),
        ("DAB", ("02", "01")),
    ]
    for gene_name, fields in examples:
        expected_gene = Gene.get("Orni", gene_name)
        expected_allele = Allele.get("Orni", gene_name, fields)
        assert expected_gene is not None
        assert expected_allele is not None
        eq_(expected_gene.mhc_class, "IIa")
        eq_(parse(f"Orni-{gene_name}", raise_on_error=True), expected_gene)
        eq_(
            parse(
                f"Orni-{gene_name}*{':'.join(fields)}",
                raise_on_error=True,
            ),
            expected_allele,
        )


def test_parse_olive_flounder_genes():
    for gene_name in ["Ia1", "Ia2", "DAA", "DAB"]:
        expected = Gene.get("Paol", gene_name)
        assert expected is not None
        eq_(parse(f"Paol-{gene_name}", raise_on_error=True), expected)


def test_parse_olive_flounder_alleles():
    for gene_name in ["Ia1", "Ia2", "DAA", "DAB"]:
        expected = Allele.get("Paol", gene_name, "01", "01")
        assert expected is not None
        eq_(parse(f"Paol-{gene_name}*01:01", raise_on_error=True), expected)


def test_parse_tongue_sole_species_Cyse():
    expected = Species.get("Cyse")
    assert expected is not None
    eq_(parse("Cyse", raise_on_error=True), expected)


def test_parse_tongue_sole_genes():
    for gene_name in ["DAA", "DAB", "DBA", "DBB"]:
        expected = Gene.get("Cyse", gene_name)
        assert expected is not None
        eq_(parse(f"Cyse-{gene_name}", raise_on_error=True), expected)


def test_parse_tongue_sole_alleles():
    for gene_name in ["DAA", "DAB", "DBA", "DBB"]:
        expected = Allele.get("Cyse", gene_name, "01", "01")
        assert expected is not None
        eq_(parse(f"Cyse-{gene_name}*01:01", raise_on_error=True), expected)


def test_parse_orange_spotted_grouper_species_Epco():
    expected = Species.get("Epco")
    assert expected is not None
    eq_(parse("Epco", raise_on_error=True), expected)


def test_parse_orange_spotted_grouper_genes():
    for gene_name in ["DAA", "DAB", "DBB"]:
        expected = Gene.get("Epco", gene_name)
        assert expected is not None
        eq_(parse(f"Epco-{gene_name}", raise_on_error=True), expected)


def test_parse_orange_spotted_grouper_alleles():
    for gene_name in ["DAA", "DAB", "DBB"]:
        expected = Allele.get("Epco", gene_name, "01", "01")
        assert expected is not None
        eq_(parse(f"Epco-{gene_name}*01:01", raise_on_error=True), expected)


def test_parse_brown_trout_species_Satr():
    expected = Species.get("Satr")
    assert expected is not None
    eq_(parse("Satr", raise_on_error=True), expected)


def test_parse_brown_trout_supported_gene_families():
    expected_classes = {"UBA": "Ia", "DAB": "IIa"}
    for gene_name in ["UBA", "DAB"]:
        expected = Gene.get("Satr", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, expected_classes[gene_name])
        eq_(parse(f"Satr-{gene_name}", raise_on_error=True), expected)


def test_parse_zebrafish_class2_genes():
    for gene_name in ["DAA", "DAB"]:
        expected = Gene.get("Dare", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, "IIa")
        eq_(parse(f"Dare-{gene_name}", raise_on_error=True), expected)


def test_parse_zebrafish_additional_class1_genes():
    for gene_name in ["UEA", "UGA"]:
        expected = Gene.get("Dare", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, "Ia")
        eq_(parse(f"Dare-{gene_name}", raise_on_error=True), expected)


def test_parse_medaka_species_Oryl():
    expected = Species.get("Oryl")
    assert expected is not None
    eq_(parse("Oryl", raise_on_error=True), expected)
    eq_(Species.get("Oryzias latipes"), expected)


def test_parse_medaka_genes():
    for gene_name in ["UAA", "UBA", "UGA", "UHA"]:
        expected = Gene.get("Oryl", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, "Ia")
        eq_(parse(f"Oryl-{gene_name}", raise_on_error=True), expected)


def test_medaka_orla_does_not_collide_with_orangutan():
    """Orla should still resolve to orangutan, not medaka."""
    result = Species.get("OrLA")
    assert result is not None
    eq_(result.name, "Pongo sp.")


def test_parse_zebrafish_uma_gene():
    expected = Gene.get("Dare", "UMA")
    assert expected is not None
    eq_(expected.mhc_class, "Ia")
    eq_(parse("Dare-UMA", raise_on_error=True), expected)


def test_parse_zebrafish_mhc1uma_alias():
    expected = Gene.get("Dare", "UMA")
    assert expected is not None
    eq_(parse("Dare-mhc1uma", raise_on_error=True), expected)


def test_parse_zebrafish_zfin_style_aliases():
    examples = [
        ("Dare-mhc1uba", Gene.get("Dare", "UBA")),
        ("Dare-mhc2daa", Gene.get("Dare", "DAA")),
        ("Dare-mhc2dab", Gene.get("Dare", "DAB")),
    ]
    for raw_string, expected in examples:
        assert expected is not None
        eq_(parse(raw_string, raise_on_error=True), expected)


def test_parse_olive_flounder_dab_numbered_loci():
    """DAB1-3 are real genes; DAB4-6 alias to DAB."""
    for alias in ["DAB1", "DAB2", "DAB3"]:
        expected = Gene.get("Paol", alias)
        assert expected is not None
        eq_(parse(f"Paol-{alias}", raise_on_error=True), expected)
    for alias in ["DAB4", "DAB5", "DAB6"]:
        expected = Gene.get("Paol", "DAB")
        assert expected is not None
        eq_(parse(f"Paol-{alias}", raise_on_error=True), expected)


def test_parse_olive_flounder_dab_alleles():
    for alias in ["DAB1", "DAB3"]:
        expected = Allele.get("Paol", alias, "01", "01")
        assert expected is not None
        eq_(parse(f"Paol-{alias}*01:01", raise_on_error=True), expected)
    expected = Allele.get("Paol", "DAB", "01", "01")
    assert expected is not None
    eq_(parse("Paol-DAB6*01:01", raise_on_error=True), expected)


def test_parse_golden_pompano_species_Trov():
    expected = Species.get("Trov")
    assert expected is not None
    eq_(parse("Trov", raise_on_error=True), expected)


def test_parse_golden_pompano_genes():
    expected_classes = {"UBA": "Ia", "DAA": "IIa", "DAB": "IIa"}
    for gene_name in ["UBA", "DAA", "DAB"]:
        expected = Gene.get("Trov", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, expected_classes[gene_name])
        eq_(parse(f"Trov-{gene_name}", raise_on_error=True), expected)


def test_parse_golden_pompano_alleles():
    for gene_name in ["UBA", "DAA", "DAB"]:
        expected = Allele.get("Trov", gene_name, "01", "01")
        assert expected is not None
        eq_(parse(f"Trov-{gene_name}*01:01", raise_on_error=True), expected)


def test_parse_arctic_char_species_Saal():
    expected = Species.get("Saal")
    assert expected is not None
    eq_(parse("Saal", raise_on_error=True), expected)


def test_parse_arctic_char_genes():
    expected_classes = {"UBA": "Ia", "UEA": "Ia", "UGA": "Ia", "DAB": "IIa"}
    for gene_name in ["UBA", "UEA", "UGA", "DAB"]:
        expected = Gene.get("Saal", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, expected_classes[gene_name])
        eq_(parse(f"Saal-{gene_name}", raise_on_error=True), expected)


def test_parse_arctic_char_case_insensitive():
    """SAAL should resolve to Saal (case-insensitive normalization)."""
    expected = Gene.get("Saal", "UBA")
    assert expected is not None
    eq_(parse("SAAL-UBA", raise_on_error=True), expected)


def test_satr_daa_parses_via_salmonid_inheritance():
    """Satr-DAA parses via Salmonidae inheritance, not via a global common-gene table."""
    result = parse("Satr-DAA", raise_on_error=True)
    assert result is not None
    eq_(result.species.name, "Salmo trutta")
